"""
Exceptions module for the scenarios package.

This module defines custom exception classes used throughout the scenarios module.
These exceptions provide specific error information for different types of errors
that can occur when working with Scenarios, ScenarioLists, and related components.
"""

import re

from ..base import BaseException


class AgentListError(BaseException):
    """
    Exception raised for errors related to AgentList operations.
    
    This exception appears to be a duplicate of the exception defined in
    edsl.agents.exceptions. It exists here for legacy reasons but is not
    actively used from this module.
    
    Note: This exception is defined but not used from this module. The AgentListError
    from edsl.agents.exceptions is used instead. This raises Exception("not used")
    to indicate this state.
    """
    
    def __init__(self, message: str):
        """
        Initialize the AgentListError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class ScenarioError(BaseException):
    """
    Exception raised for errors related to Scenario operations.
    
    This exception is raised when:
    - Invalid data is passed to initialize a Scenario (not convertible to dictionary)
    - Required fields are missing in scenario data
    - File operations fail when loading scenarios from files
    - Scenario content cannot be properly parsed or processed
    - Scenario lists encounter issues with data formats or operations
    
    To fix this error:
    1. Check that your scenario data is properly formatted (valid dictionary or convertible to one)
    2. Ensure all required fields for a scenario are present
    3. Verify file paths and permissions when loading from files
    4. Check for syntax or format errors in scenario content
    
    Examples:
        ```python
        Scenario(123)  # Raises ScenarioError (not convertible to dictionary)
        Scenario({"invalid_format": True})  # May raise ScenarioError (missing required fields)
        ```
    
    Args:
        message: A description of the error that occurred.
    """
    
    documentation = "https://docs.expectedparrot.com/en/latest/scenarios.html#module-edsl.scenarios.Scenario"

    def __init__(self, message: str):
        """
        Initialize the ScenarioError with a message and add documentation link.
        
        Args:
            message: A description of the error that occurred.
        """
        self.message = message + "\n" + "Documentation: " + self.documentation
        super().__init__(self.message)

    def __str__(self) -> str:
        """
        Return a string representation of the error with clickable URLs.
        
        This method makes any URLs in the error message clickable when displayed
        in terminal environments that support ANSI escape sequences.
        
        Returns:
            The error message with clickable URLs.
        """
        return self.make_urls_clickable(self.message)

    @staticmethod
    def make_urls_clickable(text: str) -> str:
        """
        Convert URLs in text to clickable links in terminal output.
        
        This method finds all URLs in the given text and wraps them in ANSI
        escape sequences that make them clickable in supporting terminals.
        
        Args:
            text: The text containing URLs to make clickable.
            
        Returns:
            The text with URLs converted to clickable links.
            
        Example:
            ```python
            error = ScenarioError("See docs at https://example.com")
            s = str(error)  # Returns the message with clickable link
            ```
        """
        url_pattern = r"https?://[^\s]+"
        urls = re.findall(url_pattern, text)
        for url in urls:
            clickable_url = f"\033]8;;{url}\007{url}\033]8;;\007"
            text = text.replace(url, clickable_url)
        return text


class FileNotFoundScenarioError(ScenarioError):
    """
    Exception raised when a file needed for a scenario cannot be found.
    
    This exception occurs when:
    - A file specified in a file path does not exist
    - A referenced image, document, or other resource is missing
    - A directory expected to contain scenario files is not found
    
    To fix this error:
    1. Check that the file path is correct and the file exists
    2. Verify file system permissions allow access to the file
    3. Ensure any referenced external resources are properly available
    
    Examples:
        ```python
        Scenario.from_file("/path/to/nonexistent/file.json")  # Raises FileNotFoundScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the FileNotFoundScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class ImportScenarioError(ScenarioError):
    """
    Exception raised when importing a library needed for scenario operations fails.
    
    This exception occurs when:
    - A required library for handling specific file types is not installed
    - A module needed for processing scenario data cannot be imported
    - Optional dependencies for advanced features are missing
    
    To fix this error:
    1. Install the required dependencies mentioned in the error message
    2. Check for version conflicts between dependencies
    3. Ensure your environment has all necessary packages
    
    Examples:
        ```python
        # When attempting to load a PDF without the pdf dependencies
        Scenario.from_pdf("document.pdf")  # Raises ImportScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the ImportScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class TypeScenarioError(ScenarioError):
    """
    Exception raised when there's a type mismatch in scenario operations.
    
    This exception occurs when:
    - A parameter is of the wrong type for a scenario operation
    - Incompatible types are used in scenario methods
    - Type conversion fails during scenario processing
    
    To fix this error:
    1. Check the types of parameters passed to scenario methods
    2. Ensure data structures match what scenario operations expect
    3. Verify that operations between scenarios and other objects are compatible
    
    Examples:
        ```python
        scenario * "invalid_operand"  # Raises TypeScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the TypeScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class ValueScenarioError(ScenarioError):
    """
    Exception raised when there's an invalid value in scenario operations.
    
    This exception occurs when:
    - A parameter value is out of its acceptable range
    - Invalid formats are provided for scenario data
    - Operation parameters are invalid for the requested action
    
    To fix this error:
    1. Check parameter values against allowed ranges or formats
    2. Verify inputs meet the requirements for specific operations
    3. Ensure data formats match what's expected by scenario methods
    
    Examples:
        ```python
        scenario_list.to_table(output_type="invalid_format")  # Raises ValueScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the ValueScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class AttributeScenarioError(ScenarioError):
    """
    Exception raised when accessing a non-existent attribute in a scenario.
    
    This exception occurs when:
    - Attempting to access a field not present in a scenario
    - Using an attribute accessor on a scenario for a missing property
    - CSV or dataframe column access issues
    
    To fix this error:
    1. Check that the attribute name is correct
    2. Verify the scenario contains the expected fields
    3. Use hasattr() to check for attribute existence before access
    
    Examples:
        ```python
        scenario.nonexistent_attribute  # Raises AttributeScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the AttributeScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class IndexScenarioError(ScenarioError):
    """
    Exception raised when an index is out of range in scenario operations.
    
    This exception occurs when:
    - Accessing a scenario index outside the valid range
    - Using an invalid index in a scenario list operation
    - Sequence operations with invalid indices
    
    To fix this error:
    1. Check array boundaries before accessing elements
    2. Verify indices are within valid ranges for the collection
    3. Use len() to determine the valid index range
    
    Examples:
        ```python
        scenario_list[999]  # Raises IndexScenarioError if fewer items exist
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the IndexScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)


class KeyScenarioError(ScenarioError):
    """
    Exception raised when a key is missing in scenario operations.
    
    This exception occurs when:
    - Accessing a non-existent key in a scenario
    - Using key-based access for missing fields
    - Dictionary operations with invalid keys
    
    To fix this error:
    1. Check if the key exists before attempting access
    2. Use dictionary get() method with default values for safer access
    3. Verify the scenario contains the expected keys
    
    Examples:
        ```python
        scenario["missing_key"]  # Raises KeyScenarioError
        ```
    """
    
    def __init__(self, message: str):
        """
        Initialize the KeyScenarioError with a message.
        
        Args:
            message: A description of the error that occurred.
        """
        super().__init__(message)




from enum import Enum


class JobsStatus(Enum):
    QUEUED = "queued"
    RUNNING = "running"
    COMPLETED = "completed"
    PARTIALLY_FAILED = "partially_failed"
    FAILED = "failed"
    CANCELLED = "cancelled"



from typing import Optional
import subprocess
import os
from pathlib import Path
import tempfile


class MarkdownToPDF:
    def __init__(self, markdown_content: str, filename: Optional[str] = None):
        """
        Initialize the converter with markdown content.

        Args:
            markdown_content (str): The markdown content to be converted
        """
        self.markdown_content = markdown_content
        self.filename = filename
        self.has_pandoc = self._check_pandoc()
        # self.convert()

    def _check_pandoc(self):
        """
        Check if pandoc is installed and accessible.
        
        Returns:
            bool: True if pandoc is available, False otherwise
        """
        try:
            subprocess.run(["pandoc", "--version"], capture_output=True, check=True)
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            import warnings
            warnings.warn(
                "Pandoc is not installed or not found in PATH. "
                "PDF conversion will not be available."
            )
            return False

    def convert(self, output_path: str, **options) -> bool:
        """
        Convert the markdown content to PDF.

        Args:
            output_path (str): Path where the PDF should be saved
            **options: Additional conversion options
                margin (str): Page margin (default: "1in")
                font_size (str): Font size (default: "12pt")
                toc (bool): Include table of contents (default: False)
                number_sections (bool): Number sections (default: False)
                highlight_style (str): Code highlighting style (default: "tango")

        Returns:
            bool: True if conversion was successful, False otherwise
        """
        # Ensure output directory exists
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        # Build pandoc command with default options
        cmd = [
            "pandoc",
            "-f",
            "markdown",
            "-o",
            str(output_path),
            "--pdf-engine=xelatex",
            "--variable",
            f'geometry:margin={options.get("margin", "1in")}',
            "--variable",
            f'fontsize={options.get("font_size", "12pt")}',
        ]

        # Add font only if specifically provided
        if "font" in options:
            cmd.extend(["--variable", f'mainfont={options["font"]}'])

        # Add optional parameters
        if options.get("toc", False):
            cmd.append("--toc")

        if options.get("number_sections", False):
            cmd.append("--number-sections")

        if "highlight_style" in options:
            cmd.extend(["--highlight-style", options["highlight_style"]])

        try:
            # Run pandoc command
            subprocess.run(
                cmd,
                input=self.markdown_content,
                text=True,
                capture_output=True,
                check=True,
            )
            return True
        except subprocess.CalledProcessError as e:
            print(f"Error converting markdown to PDF: {e.stderr}")
            return False

    def preview(self) -> str:
        """
        Generate a temporary PDF and return its path.

        Returns:
            str: Path to the temporary PDF file
        """
        temp_dir = tempfile.mkdtemp()
        if self.filename:
            temp_pdf = os.path.join(temp_dir, f"{self.filename}.pdf")
        else:
            temp_pdf = os.path.join(temp_dir, "preview.pdf")

        if self.convert(temp_pdf):
            from ..scenarios import FileStore

            return FileStore(temp_pdf)

        return None



from __future__ import annotations
from typing import Any, Optional, TYPE_CHECKING

import random
from jinja2 import Template
from pydantic import BaseModel, Field, model_validator, ValidationError
from typing import List, Literal, Annotated

from .exceptions import QuestionAnswerValidationError
from ..scenarios import Scenario

from .question_base import QuestionBase
from .descriptors import (
    IntegerDescriptor,
    QuestionOptionsDescriptor,
)
from .decorators import inject_exception
from .response_validator_abc import ResponseValidatorABC

if TYPE_CHECKING:
    pass


class CheckboxResponse(BaseModel):
    """
    Base Pydantic model for validating checkbox responses.

    This model defines the structure and validation rules for responses to
    checkbox questions, ensuring that selected options are properly formatted
    as a list of choices.

    Attributes:
        answer: List of selected choices
        comment: Optional comment provided with the answer
        generated_tokens: Optional raw LLM output for token tracking

    Examples:
        >>> # Valid response with list of options
        >>> response = CheckboxResponse(answer=[0, 1])
        >>> response.answer
        [0, 1]

        >>> # Valid response with comment
        >>> response = CheckboxResponse(answer=[1], comment="This is my choice")
        >>> response.answer
        [1]
        >>> response.comment
        'This is my choice'

        >>> # Invalid non-list answer
        >>> try:
        ...     CheckboxResponse(answer=1)
        ... except Exception as e:
        ...     print("Validation error occurred")
        Validation error occurred
    """

    answer: List[Any]
    comment: Optional[str] = None
    generated_tokens: Optional[Any] = None


def create_checkbox_response_model(
    choices: list,
    min_selections: Optional[int] = None,
    max_selections: Optional[int] = None,
    permissive: bool = False,
):
    """
    Dynamically create a CheckboxResponse model with a predefined list of choices.

    This function creates a customized Pydantic model for checkbox questions that
    validates both the format of the response and any constraints on selection count.

    Args:
        choices: A list of allowed values for the answer field
        min_selections: Optional minimum number of selections required
        max_selections: Optional maximum number of selections allowed
        permissive: If True, constraints are not enforced

    Returns:
        A new Pydantic model class with appropriate validation

    Examples:
        >>> # Create model with constraints
        >>> choices = [0, 1, 2, 3]
        >>> ConstrainedModel = create_checkbox_response_model(
        ...     choices=choices,
        ...     min_selections=1,
        ...     max_selections=2
        ... )

        >>> # Valid response within constraints
        >>> response = ConstrainedModel(answer=[0, 1])
        >>> response.answer
        [0, 1]

        >>> # Too few selections fails validation
        >>> try:
        ...     ConstrainedModel(answer=[])
        ... except Exception as e:
        ...     "at least 1" in str(e)
        True

        >>> # Too many selections fails validation
        >>> try:
        ...     ConstrainedModel(answer=[0, 1, 2])
        ... except Exception as e:
        ...     "at most 2" in str(e)
        True

        >>> # Invalid choice fails validation
        >>> try:
        ...     ConstrainedModel(answer=[4])
        ... except Exception as e:
        ...     any(x in str(e) for x in ["Invalid choice", "not a valid enumeration member", "validation error"])
        True

        >>> # Permissive model ignores constraints
        >>> PermissiveModel = create_checkbox_response_model(
        ...     choices=choices,
        ...     min_selections=1,
        ...     max_selections=2,
        ...     permissive=True
        ... )
        >>> response = PermissiveModel(answer=[0, 1, 2])
        >>> len(response.answer)
        3
    """
    # Convert the choices list to a tuple for use with Literal
    choice_tuple = tuple(choices)

    if permissive:
        # For permissive mode, we still validate the choice values but ignore count constraints
        class PermissiveCheckboxResponse(CheckboxResponse):
            """Checkbox response model with choices validation but no count constraints."""

            answer: Annotated[
                List[Literal[choice_tuple]],
                Field(description="List of selected choices"),
            ]

            @model_validator(mode="after")
            def validate_choices(self):
                """Validate that each selected choice is valid."""
                for choice in self.answer:
                    if choice not in choices:
                        validation_error = ValidationError.from_exception_data(
                            title="CheckboxResponse",
                            line_errors=[
                                {
                                    "type": "value_error",
                                    "loc": ("answer",),
                                    "msg": f"Invalid choice: {choice}. Must be one of: {choices}",
                                    "input": choice,
                                    "ctx": {"error": "Invalid choice"},
                                }
                            ],
                        )
                        raise QuestionAnswerValidationError(
                            message=f"Invalid choice: {choice}. Must be one of: {choices}",
                            data=self.model_dump(),
                            model=self.__class__,
                            pydantic_error=validation_error,
                        )
                return self

        return PermissiveCheckboxResponse
    else:
        # For non-permissive mode, enforce both choice values and count constraints
        class ConstrainedCheckboxResponse(CheckboxResponse):
            """Checkbox response model with both choice and count constraints."""

            answer: Annotated[
                List[Literal[choice_tuple]],
                Field(description="List of selected choices"),
            ]

            @model_validator(mode="after")
            def validate_selection_count(self):
                """Validate that the number of selections meets constraints."""
                if min_selections is not None and len(self.answer) < min_selections:
                    validation_error = ValidationError.from_exception_data(
                        title="CheckboxResponse",
                        line_errors=[
                            {
                                "type": "value_error",
                                "loc": ("answer",),
                                "msg": f"Must select at least {min_selections} option(s)",
                                "input": self.answer,
                                "ctx": {"error": "Too few selections"},
                            }
                        ],
                    )
                    raise QuestionAnswerValidationError(
                        message=f"Must select at least {min_selections} option(s), got {len(self.answer)}",
                        data=self.model_dump(),
                        model=self.__class__,
                        pydantic_error=validation_error,
                    )

                if max_selections is not None and len(self.answer) > max_selections:
                    validation_error = ValidationError.from_exception_data(
                        title="CheckboxResponse",
                        line_errors=[
                            {
                                "type": "value_error",
                                "loc": ("answer",),
                                "msg": f"Must select at most {max_selections} option(s)",
                                "input": self.answer,
                                "ctx": {"error": "Too many selections"},
                            }
                        ],
                    )
                    raise QuestionAnswerValidationError(
                        message=f"Must select at most {max_selections} option(s), got {len(self.answer)}",
                        data=self.model_dump(),
                        model=self.__class__,
                        pydantic_error=validation_error,
                    )

                # Also validate that each choice is valid
                for choice in self.answer:
                    if choice not in choices:
                        validation_error = ValidationError.from_exception_data(
                            title="CheckboxResponse",
                            line_errors=[
                                {
                                    "type": "value_error",
                                    "loc": ("answer",),
                                    "msg": f"Invalid choice: {choice}. Must be one of: {choices}",
                                    "input": choice,
                                    "ctx": {"error": "Invalid choice"},
                                }
                            ],
                        )
                        raise QuestionAnswerValidationError(
                            message=f"Invalid choice: {choice}. Must be one of: {choices}",
                            data=self.model_dump(),
                            model=self.__class__,
                            pydantic_error=validation_error,
                        )

                return self

        return ConstrainedCheckboxResponse


class CheckBoxResponseValidator(ResponseValidatorABC):
    """
    Validator for checkbox question responses.

    This class implements the validation and fixing logic for checkbox responses.
    It ensures that responses contain valid selections from the available options
    and that the number of selections meets any constraints.

    Attributes:
        required_params: List of required parameters for validation.
        valid_examples: Examples of valid responses for testing.
        invalid_examples: Examples of invalid responses for testing.

    Examples:
        >>> from edsl import QuestionCheckBox
        >>> q = QuestionCheckBox.example(use_code=True)
        >>> validator = q.response_validator

        >>> # Fix string to list
        >>> response = {"answer": 1}
        >>> fixed = validator.fix(response)
        >>> isinstance(fixed["answer"], list)
        True

        >>> # Extract selections from text
        >>> response = {"generated_tokens": "I choose options 0 and 2"}
        >>> fixed = validator.fix(response)
        >>> sorted(fixed["answer"])
        [0, 2]

        >>> # Fix comma-separated list
        >>> response = {"generated_tokens": "0, 1, 3"}
        >>> fixed = validator.fix(response)
        >>> sorted(fixed["answer"])
        [0, 1, 3]

        >>> # Preserve comments when fixing
        >>> response = {"answer": 1, "comment": "My explanation"}
        >>> fixed = validator.fix(response)
        >>> "comment" in fixed and fixed["comment"] == "My explanation"
        True
    """

    required_params = [
        "question_options",
        "min_selections",
        "max_selections",
        "use_code",
        "permissive",
    ]

    valid_examples = [
        ({"answer": [1, 2]}, {"question_options": ["Good", "Great", "OK", "Bad"]})
    ]

    invalid_examples = [
        (
            {"answer": [-1]},
            {"question_options": ["Good", "Great", "OK", "Bad"]},
            "Invalid choice",
        ),
        (
            {"answer": 1},
            {"question_options": ["Good", "Great", "OK", "Bad"]},
            "value is not a valid list",
        ),
        (
            {"answer": [1, 2, 3, 4]},
            {
                "question_options": ["Good", "Great", "OK", "Bad"],
                "min_selections": 1,
                "max_selections": 2,
            },
            "Must select at most 2",
        ),
    ]

    def fix(self, response, verbose=False):
        """
        Fix common issues in checkbox responses.

        This method attempts to extract valid selections from responses with
        format issues. It can handle:
        1. Single values that should be lists
        2. Comma-separated strings in answer field or generated_tokens
        3. Finding option indices mentioned in text

        Args:
            response: The response dictionary to fix
            verbose: If True, print information about the fixing process

        Returns:
            A fixed version of the response dictionary with a valid list of selections

        Notes:
            - First tries to convert to a list if the answer is not already a list
            - Then tries to parse comma-separated values from answer or generated_tokens
            - Finally tries to find option indices mentioned in the text
            - Preserves any comment in the original response
        """
        if verbose:
            print("Invalid response of QuestionCheckBox was: ", response)

        # Check if answer exists and is a comma-separated string (common LLM output format)
        if (
            "answer" in response
            and isinstance(response["answer"], str)
            and "," in response["answer"]
        ):
            if verbose:
                print(f"Parsing comma-separated answer string: {response['answer']}")

            # Split by commas and strip whitespace
            proposed_list = response["answer"].split(",")
            proposed_list = [item.strip() for item in proposed_list]

            # Try to convert to integers if use_code is True
            if self.use_code:
                try:
                    proposed_list = [int(i) for i in proposed_list]
                except ValueError:
                    # If we can't convert to integers, try to match values to indices
                    if verbose:
                        print(
                            "Could not convert comma-separated values to integers, trying to match options"
                        )

                    # Try to match option text values to their indices
                    index_map = {}
                    for i, option in enumerate(self.question_options):
                        index_map[option.lower().strip()] = i

                    converted_list = []
                    for item in proposed_list:
                        item_lower = item.lower().strip()
                        if item_lower in index_map:
                            converted_list.append(index_map[item_lower])

                    if converted_list:
                        proposed_list = converted_list

            if verbose:
                print("Proposed solution from comma separation is: ", proposed_list)

            proposed_data = {
                "answer": proposed_list,
                "comment": response.get("comment"),
                "generated_tokens": response.get("generated_tokens"),
            }

            # Try validating with the proposed solution
            try:
                validated = self._base_validate(proposed_data)
                return validated.model_dump()
            except Exception as e:
                if verbose:
                    print(f"Comma-separated solution invalid: {e}")

        # If answer exists but is not a list, convert it to a list
        elif "answer" in response and not isinstance(response["answer"], list):
            if verbose:
                print(f"Converting non-list answer {response['answer']} to a list")
            answer_value = response["answer"]
            response = {**response, "answer": [answer_value]}

            # Try validating the fixed response
            try:
                validated = self._base_validate(response)
                return validated.model_dump()
            except Exception:
                if verbose:
                    print("Converting to list didn't fix the issue")

        # Try parsing from generated_tokens if present
        response_text = response.get("generated_tokens")
        if response_text and isinstance(response_text, str):
            # Try comma-separated list first
            if "," in response_text:
                proposed_list = response_text.split(",")
                proposed_list = [item.strip() for item in proposed_list]

                if self.use_code:
                    try:
                        proposed_list = [int(i) for i in proposed_list]
                    except ValueError:
                        # If we can't convert to integers, try to match values to indices
                        if verbose:
                            print(
                                "Could not convert comma-separated values to integers, trying to match options"
                            )

                        # Try to match option text values to their indices
                        index_map = {}
                        for i, option in enumerate(self.question_options):
                            index_map[option.lower().strip()] = i

                        converted_list = []
                        for item in proposed_list:
                            item_lower = item.lower().strip()
                            if item_lower in index_map:
                                converted_list.append(index_map[item_lower])

                        if converted_list:
                            proposed_list = converted_list

                if verbose:
                    print("Proposed solution from comma separation is: ", proposed_list)

                proposed_data = {
                    "answer": proposed_list,
                    "comment": response.get("comment"),
                    "generated_tokens": response.get("generated_tokens"),
                }

                # Try validating with the proposed solution
                try:
                    validated = self._base_validate(proposed_data)
                    return validated.model_dump()
                except Exception as e:
                    if verbose:
                        print(f"Comma-separated solution invalid: {e}")

            # Try finding option indices mentioned in the text
            matches = []
            for index, option in enumerate(self.question_options):
                if self.use_code:
                    if str(index) in response_text:
                        matches.append(index)
                else:
                    if option in response_text:
                        matches.append(option)

            if matches:
                if verbose:
                    print(f"Found options mentioned in text: {matches}")

                proposed_data = {
                    "answer": matches,
                    "comment": response.get("comment"),
                    "generated_tokens": response.get("generated_tokens"),
                }

                # Try validating with the proposed solution
                try:
                    validated = self._base_validate(proposed_data)
                    return validated.model_dump()
                except Exception as e:
                    if verbose:
                        print(f"Text matching solution invalid: {e}")

        # If nothing worked, return the original response
        return response


class QuestionCheckBox(QuestionBase):
    """
    A question that prompts the agent to select multiple options from a list.

    QuestionCheckBox allows agents to select one or more items from a predefined
    list of options. It's useful for "select all that apply" scenarios, multi-select
    preferences, or any question where multiple valid selections are possible.

    Attributes:
        question_type (str): Identifier for this question type, set to "checkbox".
        purpose (str): Brief description of when to use this question type.
        question_options: List of available options to select from.
        min_selections: Optional minimum number of selections required.
        max_selections: Optional maximum number of selections allowed.
        _response_model: Initially None, set by create_response_model().
        response_validator_class: Class used to validate and fix responses.

    Examples:
        >>> # Basic creation works
        >>> q = QuestionCheckBox.example()
        >>> q.question_type
        'checkbox'

        >>> # Create preferences question with selection constraints
        >>> q = QuestionCheckBox(
        ...     question_name="favorite_fruits",
        ...     question_text="Which fruits do you like?",
        ...     question_options=["Apple", "Banana", "Cherry", "Durian", "Elderberry"],
        ...     min_selections=1,
        ...     max_selections=3
        ... )
        >>> q.question_options
        ['Apple', 'Banana', 'Cherry', 'Durian', 'Elderberry']
        >>> q.min_selections
        1
        >>> q.max_selections
        3
    """

    question_type = "checkbox"
    purpose = "When options are known and limited"
    question_options: list[str] = QuestionOptionsDescriptor()
    min_selections = IntegerDescriptor(none_allowed=True)
    max_selections = IntegerDescriptor(none_allowed=True)

    _response_model = None
    response_validator_class = CheckBoxResponseValidator

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        min_selections: Optional[int] = None,
        max_selections: Optional[int] = None,
        include_comment: bool = True,
        use_code: bool = False,
        question_presentation: Optional[str] = None,
        answering_instructions: Optional[str] = None,
        permissive: bool = False,
    ):
        """
        Initialize a new checkbox question.

        Args:
            question_name: Identifier for the question, used in results and templates.
                          Must be a valid Python variable name.
            question_text: The actual text of the question to be asked.
            question_options: List of options the agent can select from.
            min_selections: Optional minimum number of options that must be selected.
            max_selections: Optional maximum number of options that can be selected.
            include_comment: Whether to allow comments with the answer.
            use_code: If True, use indices (0,1,2...) instead of option text values.
            question_presentation: Optional custom presentation template.
            answering_instructions: Optional additional instructions.
            permissive: If True, ignore selection count constraints during validation.

        Examples:
            >>> q = QuestionCheckBox(
            ...     question_name="symptoms",
            ...     question_text="Select all symptoms you are experiencing:",
            ...     question_options=["Fever", "Cough", "Headache", "Fatigue"],
            ...     min_selections=1
            ... )
            >>> q.question_name
            'symptoms'

            >>> # Question with both min and max
            >>> q = QuestionCheckBox(
            ...     question_name="pizza_toppings",
            ...     question_text="Choose 2-4 toppings for your pizza:",
            ...     question_options=["Cheese", "Pepperoni", "Mushroom", "Onion",
            ...                       "Sausage", "Bacon", "Pineapple"],
            ...     min_selections=2,
            ...     max_selections=4
            ... )
        """
        self.question_name = question_name
        self.question_text = question_text
        self.min_selections = min_selections
        self.max_selections = max_selections
        self.question_options = question_options

        self._include_comment = include_comment
        self._use_code = use_code
        self.permissive = permissive

        self.question_presentation = question_presentation
        self.answering_instructions = answering_instructions

    def create_response_model(self):
        """
        Create a response model with the appropriate constraints.

        This method creates a Pydantic model customized with the options and
        selection count constraints specified for this question instance.

        Returns:
            A Pydantic model class tailored to this question's constraints.

        Examples:
            >>> q = QuestionCheckBox.example(use_code=True)
            >>> model = q.create_response_model()
            >>> model(answer=[0, 2])  # Select first and third options
            ConstrainedCheckboxResponse(answer=[0, 2], comment=None, generated_tokens=None)
        """
        if not self._use_code:
            # Use option text values as valid choices
            return create_checkbox_response_model(
                self.question_options,
                min_selections=self.min_selections,
                max_selections=self.max_selections,
                permissive=self.permissive,
            )
        else:
            # Use option indices (0, 1, 2...) as valid choices
            return create_checkbox_response_model(
                list(range(len(self.question_options))),
                min_selections=self.min_selections,
                max_selections=self.max_selections,
                permissive=self.permissive,
            )

    def _translate_answer_code_to_answer(
        self, answer_codes, scenario: "Scenario" = None
    ):
        """
        Translate the answer codes to the actual answer text.

        For checkbox questions with use_code=True, the agent responds with
        option indices (e.g., [0, 1]) which need to be translated to their
        corresponding option text values (e.g., ["Option A", "Option B"]).

        Args:
            answer_codes: List of selected option indices or values
            scenario: Optional scenario with variables for template rendering

        Returns:
            List of selected option texts

        Examples:
            >>> q = QuestionCheckBox(
            ...     question_name="example",
            ...     question_text="Select options:",
            ...     question_options=["A", "B", "C"],use_code=True
            ... )
            >>> q._translate_answer_code_to_answer([0, 2])
            ['A', 'C']
        """
        scenario = scenario or Scenario()
        translated_options = [
            Template(str(option)).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            if self._use_code:
                translated_codes.append(translated_options[int(answer_code)])
            else:
                translated_codes.append(answer_code)
        return translated_codes

    def _simulate_answer(self, human_readable=True):
        """
        Simulate a valid answer for debugging purposes.

        This method generates a random valid answer for the checkbox question,
        useful for testing and demonstrations.

        Args:
            human_readable: If True, return option text values; if False, return indices

        Returns:
            A dictionary with a valid random answer

        Examples:
            >>> q = QuestionCheckBox.example()
            >>> answer = q._simulate_answer(human_readable=False)
            >>> len(answer["answer"]) >= q.min_selections
            True
            >>> len(answer["answer"]) <= q.max_selections
            True
        """
        from edsl.utilities.utilities import random_string

        min_sel = self.min_selections or 1
        max_sel = self.max_selections or len(self.question_options)
        # Ensure we don't try to select more options than available
        max_sel = min(max_sel, len(self.question_options))
        min_sel = min(min_sel, max_sel)

        num_selections = random.randint(min_sel, max_sel)

        if human_readable:
            # Select a random number of options from self.question_options
            selected_options = random.sample(self.question_options, num_selections)
            answer = {
                "answer": selected_options,
                "comment": random_string(),
            }
        else:
            # Select a random number of indices from the range of self.question_options
            selected_indices = random.sample(
                range(len(self.question_options)), num_selections
            )
            answer = {
                "answer": selected_indices,
                "comment": random_string(),
            }
        return answer

    @property
    def question_html_content(self) -> str:
        """
        Generate HTML content for rendering the question in web interfaces.

        This property generates HTML markup for the question when it needs to be
        displayed in web interfaces or HTML contexts. For a checkbox question,
        this is a set of checkbox input elements, one for each option.

        Returns:
            str: HTML markup for rendering the question.
        """
        instructions = ""
        if self.min_selections is not None:
            instructions += f"Select at least {self.min_selections} option(s). "
        if self.max_selections is not None:
            instructions += f"Select at most {self.max_selections} option(s)."

        question_html_content = Template(
            """
        <p>{{ instructions }}</p>
        {% for option in question_options %} 
        <div>
        <input type="checkbox" id="{{ option }}" name="{{ question_name }}" value="{{ option }}">
        <label for="{{ option }}">{{ option }}</label>
        </div>
        {% endfor %}
        """
        ).render(
            instructions=instructions,
            question_name=self.question_name,
            question_options=self.question_options,
        )
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    @inject_exception
    def example(cls, include_comment=False, use_code=False) -> QuestionCheckBox:
        """
        Create an example instance of a checkbox question.

        This class method creates a predefined example of a checkbox question
        for demonstration, testing, and documentation purposes.

        Args:
            include_comment: Whether to include a comment field with the answer.
            use_code: Whether to use indices (True) or values (False) for answer codes.

        Returns:
            QuestionCheckBox: An example checkbox question.

        Examples:
            >>> q = QuestionCheckBox.example()
            >>> q.question_name
            'never_eat'
            >>> len(q.question_options)
            5
            >>> q.min_selections
            2
            >>> q.max_selections
            5
        """
        return cls(
            question_name="never_eat",
            question_text="Which of the following foods would you eat if you had to?",
            question_options=[
                "soggy meatpie",
                "rare snails",
                "mouldy bread",
                "panda milk custard",
                "McDonalds",
            ],
            min_selections=2,
            max_selections=5,
            use_code=use_code,
            include_comment=include_comment,
        )


def main():
    """
    Demonstrate the functionality of the QuestionCheckBox class.

    This function creates an example checkbox question and demonstrates its
    key features including validation, serialization, and answer simulation.
    It's primarily intended for testing and development purposes.

    Note:
        This function will be executed when the module is run directly,
        but not when imported.
    """
    print("Creating a QuestionCheckBox example...")
    q = QuestionCheckBox.example()
    print(f"Question text: {q.question_text}")
    print(f"Question name: {q.question_name}")
    print(f"Question options: {q.question_options}")
    print(f"Min selections: {q.min_selections}")
    print(f"Max selections: {q.max_selections}")

    # Validate an answer
    print("\nValidating an answer...")
    valid_answer = {"answer": [1, 2], "comment": "I like these foods"}
    validated = q._validate_answer(valid_answer)
    print(f"Validated answer: {validated}")

    # Translate answer codes
    print("\nTranslating answer codes...")
    translated = q._translate_answer_code_to_answer([1, 2])
    print(f"Translated answer: {translated}")

    # Simulate answers
    print("\nSimulating answers...")
    simulated_human = q._simulate_answer(human_readable=True)
    print(f"Simulated human-readable answer: {simulated_human}")

    simulated_codes = q._simulate_answer(human_readable=False)
    print(f"Simulated code answer: {simulated_codes}")

    # Validate simulated answer
    validated_simulated = q._validate_answer(simulated_codes)
    print(f"Validated simulated answer: {validated_simulated}")

    # Serialization demonstration
    print("\nTesting serialization...")
    serialized = q.to_dict()
    print(f"Serialized question (keys): {list(serialized.keys())}")
    deserialized = QuestionBase.from_dict(serialized)
    print(
        f"Deserialization successful: {deserialized.question_text == q.question_text}"
    )

    # Run doctests
    print("\nRunning doctests...")
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    print("Doctests completed")


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    # Uncomment to run demonstration
    # main()



import re
import uuid
from datetime import datetime
from typing import Union

from IPython.display import display, HTML

from .jobs_remote_inference_logger import JobLogger
from .jobs_status_enums import JobsStatus


class HTMLTableJobLogger(JobLogger):
    def __init__(self, verbose=True, **kwargs):
        super().__init__(verbose=verbose)
        self.display_handle = display(HTML(""), display_id=True)
        self.current_message = None
        self.log_id = str(uuid.uuid4())
        self.is_expanded = True
        self.spinner_chars = ["◐", "◓", "◑", "◒"]
        self.spinner_idx = 0
        self.messages = []  # Store message history
        self.external_link_icon = """
            <svg
                xmlns="http://www.w3.org/2000/svg"
                width="24"
                height="24"
                viewBox="0 0 24 24"
                fill="none"
                stroke="currentColor"
                stroke-width="2"
                stroke-linecap="round"
                stroke-linejoin="round"
                class="external-link-icon"
            >
                <path d="M15 3h6v6" />
                <path d="M10 14 21 3" />
                <path d="M18 13v6a2 2 0 0 1-2 2H5a2 2 0 0 1-2-2V8a2 2 0 0 1 2-2h6" />
            </svg>
        """

    def _get_status_icon(self, status: JobsStatus) -> str:
        """Return appropriate icon for job status"""
        if status == JobsStatus.RUNNING:
            spinner = self.spinner_chars[self.spinner_idx]
            self.spinner_idx = (self.spinner_idx + 1) % len(self.spinner_chars)
            return f'<span class="status-icon status-running">{spinner}</span>'
        elif status == JobsStatus.COMPLETED:
            return '<span class="status-icon status-completed">✓</span>'
        elif status == JobsStatus.PARTIALLY_FAILED:
            return '<span class="status-icon status-partially-failed">✗</span>'
        elif status == JobsStatus.FAILED:
            return '<span class="status-icon status-failed">✗</span>'
        else:
            return '<span class="status-icon status-unknown">•</span>'

    def _linkify(self, text: str) -> str:
        """Convert markdown-style links to HTML links"""
        markdown_pattern = r"\[(.*?)\]\((.*?)\)"
        return re.sub(
            markdown_pattern,
            r'<a href="\2" target="_blank" class="link">\1</a>',
            text,
        )

    def _create_uuid_copy_button(
        self, uuid_value: str, helper_text: Union[str, None] = None
    ) -> str:
        """Create a UUID display with click-to-copy functionality"""
        short_uuid = uuid_value
        if len(uuid_value) > 12:
            short_uuid = f"{uuid_value[:8]}...{uuid_value[-4:]}"

        return f"""
        <div class="uuid-container-wrapper">
            <div class="uuid-container" title="{uuid_value}">
                <span class="uuid-code">{short_uuid}</span>
                {self._create_copy_button(uuid_value)}
            </div>
            {f'<div class="helper-text">{helper_text}</div>' if helper_text else ''}
        </div>
        """

    def _create_copy_button(self, value: str) -> str:
        """Create a button with click-to-copy functionality"""
        return f"""
            <button class="copy-btn" onclick="navigator.clipboard.writeText('{value}').then(() => {{
                const btn = this;
                btn.querySelector('.copy-icon').style.display = 'none';
                btn.querySelector('.check-icon').style.display = 'block';
                setTimeout(() => {{
                    btn.querySelector('.check-icon').style.display = 'none';
                    btn.querySelector('.copy-icon').style.display = 'block';
                }}, 1000);
            }})" title="Copy to clipboard">
                <svg class="copy-icon" xmlns="http://www.w3.org/2000/svg" width="14" height="14" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round">
                    <rect x="9" y="9" width="13" height="13" rx="2" ry="2"></rect>
                    <path d="M5 15H4a2 2 0 0 1-2-2V4a2 2 0 0 1 2-2h9a2 2 0 0 1 2 2v1"></path>
                </svg>
                <svg class="check-icon" style="display: none" xmlns="http://www.w3.org/2000/svg" width="14" height="14" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round">
                    <polyline points="20 6 9 17 4 12"></polyline>
                </svg>
            </button>
        """

    def update(self, message: str, status: JobsStatus = JobsStatus.RUNNING):
        """Update the display with new message and current JobsInfo state"""
        self.current_message = message
        # Add to message history with timestamp
        self.messages.append(
            {
                "text": message,
                "status": status,
                "timestamp": datetime.now().strftime("%H:%M:%S"),
            }
        )

        if self.verbose:
            self.display_handle.update(HTML(self._get_html(status)))
        else:
            return None

    def _collapse(self, content_id: str, arrow_id: str) -> str:
        """Generate the onclick JavaScript for collapsible sections"""
        return f"""
            const content = document.getElementById('{content_id}');
            const arrow = document.getElementById('{arrow_id}');
            if (content.style.display === 'none') {{
                content.style.display = 'block';
                arrow.innerHTML = '&#8963;';
            }} else {{
                content.style.display = 'none';
                arrow.innerHTML = '&#8964;';
            }}
        """

    def _build_exceptions_table(self) -> str:
        """Generate HTML for the exceptions summary table section."""
        if not self.jobs_info.exception_summary:
            return ""

        total_exceptions = sum(
            exc.exception_count for exc in self.jobs_info.exception_summary
        )

        # Generate exception rows HTML before the return
        exception_rows = "".join(
            f"""
            <tr>
                <td>{exc.exception_type or '-'}</td>
                <td>{exc.inference_service or '-'}</td>
                <td>{exc.model or '-'}</td>
                <td>{exc.question_name or '-'}</td>
                <td class='exception-count'>{exc.exception_count:,}</td>
            </tr>
        """
            for exc in self.jobs_info.exception_summary
        )

        # Get the error report URL if it exists
        error_report_url = getattr(self.jobs_info, "error_report_url", None)
        error_report_link = (
            f"""
            <div style="margin-bottom: 12px; font-size: 0.85em;">
                <a href="{error_report_url}" target="_blank" class="pill-link">
                    View full exceptions report{self.external_link_icon}
                </a>
            </div>
            """
            if error_report_url
            else ""
        )

        return f"""
        <div class="exception-section">
            <div class="exception-header" onclick="{self._collapse(f'exception-content-{self.log_id}', f'exception-arrow-{self.log_id}')}">
                <span id="exception-arrow-{self.log_id}" class="expand-toggle">&#8963;</span>
                <span>Exception Summary ({total_exceptions:,} total)</span>
                <span style="flex-grow: 1;"></span>
            </div>
            <div id="exception-content-{self.log_id}" class="exception-content">
                {error_report_link}
                <table class='exception-table'>
                    <thead>
                        <tr>
                            <th>Exception Type</th>
                            <th>Service</th>
                            <th>Model</th>
                            <th>Question</th>
                            <th>Count</th>
                        </tr>
                    </thead>
                    <tbody>
                        {exception_rows}
                    </tbody>
                </table>
            </div>
        </div>
        """

    def _build_model_costs_table(self) -> str:
        """Generate HTML for the model costs summary table section."""
        if not hasattr(self.jobs_info, "model_costs") or not self.jobs_info.model_costs:
            return ""

        # Calculate totals
        total_input_tokens = sum(
            cost.input_tokens or 0 for cost in self.jobs_info.model_costs
        )
        total_output_tokens = sum(
            cost.output_tokens or 0 for cost in self.jobs_info.model_costs
        )
        total_input_cost = sum(
            cost.input_cost_usd or 0 for cost in self.jobs_info.model_costs
        )
        total_output_cost = sum(
            cost.output_cost_usd or 0 for cost in self.jobs_info.model_costs
        )
        total_cost = total_input_cost + total_output_cost

        # Calculate credit totals
        total_input_credits = sum(
            cost.input_cost_credits_with_cache or 0
            for cost in self.jobs_info.model_costs
        )
        total_output_credits = sum(
            cost.output_cost_credits_with_cache or 0
            for cost in self.jobs_info.model_costs
        )
        total_credits = total_input_credits + total_output_credits

        # Generate cost rows HTML with class names for right alignment
        cost_rows = "".join(
            f"""
            <tr>
                <td>{cost.service or '-'}</td>
                <td>{cost.model or '-'}</td>
                <td class='token-count'>{cost.input_tokens:,}</td>
                <td class='cost-value'>${cost.input_cost_usd:.4f}</td>
                <td class='token-count'>{cost.output_tokens:,}</td>
                <td class='cost-value'>${cost.output_cost_usd:.4f}</td>
                <td class='cost-value'>${(cost.input_cost_usd or 0) + (cost.output_cost_usd or 0):.4f}</td>
                <td class='cost-value'>{(cost.input_cost_credits_with_cache or 0) + (cost.output_cost_credits_with_cache or 0):,.2f}</td>
            </tr>
            """
            for cost in self.jobs_info.model_costs
        )

        # Add total row with the same alignment classes
        total_row = f"""
            <tr class='totals-row'>
                <td colspan='2'><strong>Totals</strong></td>
                <td class='token-count'>{total_input_tokens:,}</td>
                <td class='cost-value'>${total_input_cost:.4f}</td>
                <td class='token-count'>{total_output_tokens:,}</td>
                <td class='cost-value'>${total_output_cost:.4f}</td>
                <td class='cost-value'>${total_cost:.4f}</td>
                <td class='cost-value'>{total_credits:,.2f}</td>
            </tr>
        """

        return f"""
        <div class="model-costs-section">
            <div class="model-costs-header" onclick="{self._collapse(f'model-costs-content-{self.log_id}', f'model-costs-arrow-{self.log_id}')}">
                <span id="model-costs-arrow-{self.log_id}" class="expand-toggle">&#8963;</span>
                <span>Model Costs (${total_cost:.4f} / {total_credits:,.2f} credits total)</span>
                <span style="flex-grow: 1;"></span>
            </div>
            <div id="model-costs-content-{self.log_id}" class="model-costs-content">
                <table class='model-costs-table'>
                    <thead>
                        <tr>
                            <th>Service</th>
                            <th>Model</th>
                            <th class="cost-header">Input Tokens</th>
                            <th class="cost-header">Input Cost</th>
                            <th class="cost-header">Output Tokens</th>
                            <th class="cost-header">Output Cost</th>
                            <th class="cost-header">Total Cost</th>
                            <th class="cost-header">Total Credits</th>
                        </tr>
                    </thead>
                    <tbody>
                        {cost_rows}
                        {total_row}
                    </tbody>
                </table>
                <p style="font-style: italic; margin-top: 8px; font-size: 0.85em; color: #4b5563;">
                    You can obtain the total credit cost by multiplying the total USD cost by 100. A lower credit cost indicates that you saved money by retrieving responses from the universal remote cache.
                </p>
            </div>
        </div>
        """

    def _get_html(self, current_status: JobsStatus = JobsStatus.RUNNING) -> str:
        """Generate the complete HTML display with modern design"""
        css = """
        <style>
            .jobs-container {
                font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen, Ubuntu, Cantarell, "Open Sans", "Helvetica Neue", sans-serif;
                max-width: 800px;
                margin: 16px 0;
                border-radius: 8px;
                overflow: hidden;
                box-shadow: 0 1px 3px rgba(0,0,0,0.12), 0 1px 2px rgba(0,0,0,0.24);
                color: #1a1a1a;
            }
            .jobs-header {
                padding: 8px 12px;
                background: linear-gradient(to right, #f7f9fc, #edf2f7);
                border-bottom: 1px solid #e2e8f0;
                cursor: pointer;
                display: flex;
                align-items: center;
                justify-content: space-between;
                font-weight: 500;
                font-size: 0.9em;
                flex-wrap: wrap;
                gap: 8px;
            }
            .jobs-content {
                background: white;
            }
            .jobs-table {
                width: 100%;
                border-collapse: collapse;
            }
            .jobs-table th, .jobs-table td {
                padding: 12px 16px;
                text-align: left;
                border-bottom: 1px solid #edf2f7;
            }
            .jobs-table th {
                width: 25%;
                max-width: 140px;
                background-color: #f7fafc;
                font-weight: 500;
                color: #4a5568;
            }
            .jobs-table td {
                width: 75%;
            }
            .section-header {
                padding: 5px 12px;
                font-weight: 600;
                color: #1f2937;
                background-color: #f7fafc;
                border-bottom: 1px solid #cbd5e1;
                font-size: 0.85em;
            }
            .three-column-grid {
                display: flex;
                flex-wrap: wrap;
                gap: 1px;
                background-color: #e2e8f0;
            }
            .column {
                background-color: white;
            }
            .column:nth-child(1) {  /* Job Links */
                flex: 1;
                min-width: 150px;
            }
            .column:nth-child(2) {  /* Content */
                flex: 1;
                min-width: 150px;
            }
            .column:nth-child(3) {  /* Identifiers */
                flex: 2;
                min-width: 300px;
            }
            .content-box {
                padding: 5px 12px;
            }
            .link-item {
                padding: 3px 0;
                border-bottom: 1px solid #f1f5f9;
                font-size: 0.9em;
            }
            .link-item:last-child {
                border-bottom: none;
            }
            .results-link .pill-link {
                color: #059669;
                font-weight: 500;
            }
            .results-link .pill-link:hover {
                border-bottom-color: #059669;
            }
            .progress-link .pill-link {
                color: #3b82f6;
                font-weight: 500;
            }
            .progress-link .pill-link:hover {
                border-bottom-color: #3b82f6;
            }
            .remote-link .pill-link {
                color: #4b5563;
                font-weight: 500;
            }
            .remote-link .pill-link:hover {
                border-bottom-color: #4b5563;
            }
            .uuid-item {
                padding: 3px 0;
                border-bottom: 1px solid #f1f5f9;
                display: flex;
                align-items: flex-start;
            }
            .uuid-item:last-child {
                border-bottom: none;
            }
            .uuid-label {
                font-weight: 500;
                color: #4b5563;
                font-size: 0.75em;
                margin-right: 6px;
                min-width: 80px;
            }
            .compact-links {
                padding: 8px 16px;
                line-height: 1.5;
            }
            .pill-link {
                font-weight: 500;
                text-decoration: none;
                border-bottom: 1px dotted #bfdbfe;
                transition: border-color 0.2s;
                display: inline-flex;
                align-items: center;
                gap: 4px;
            }
            .pill-link:hover {
                border-bottom: 1px solid #3b82f6;
            }
            .external-link-icon {
                width: 12px;
                height: 12px;
                opacity: 0.7;
            }
            .status-banner {
                display: flex;
                align-items: center;
                flex-wrap: wrap;
                gap: 8px;
                padding: 5px 12px;
                background-color: #f7fafc;
                border-top: 1px solid #edf2f7;
                font-size: 0.85em;
                cursor: pointer;
            }
            .status-running { color: #3b82f6; }
            .status-completed { color: #059669; }
            .status-partially-failed { color: #d97706; }
            .status-failed { color: #ef4444; }
            .status-unknown { color: #6b7280; }
            .status-icon {
                display: inline-flex;
                align-items: center;
                justify-content: center;
                width: 18px;
                height: 18px;
                margin-right: 6px;
                font-weight: bold;
            }
            .link {
                color: #3b82f6;
                text-decoration: none;
                border-bottom: 1px dotted #bfdbfe;
                transition: border-color 0.2s;
            }
            .link:hover {
                border-bottom: 1px solid #3b82f6;
            }
            .uuid-container-wrapper {
                display: flex;
                flex-direction: column;
                align-items: stretch;
                gap: 4px;
                flex: 1;
                padding-bottom: 4px;
            }            
            .uuid-container {
                display: flex;
                align-items: center;
                background-color: #f8fafc;
                border-radius: 3px;
                padding: 2px 6px;
                font-family: "SF Mono", "Cascadia Mono", monospace;
                font-size: 0.75em;
                width: 100%;  /* Make sure it fills the width */
            }
            .uuid-code {
                color: #4b5563;
                overflow: hidden;
                text-overflow: ellipsis;
                flex: 1;
            }
            .copy-btn {
                background-color: #e2e8f0;
                border: none;
                cursor: pointer;
                margin-left: 4px;
                padding: 4px;
                border-radius: 3px;
                transition: all 0.2s ease;
                color: #4b5563;
                display: flex;
                align-items: center;
                justify-content: center;
                width: 22px;
                height: 22px;
            }
            .copy-btn:hover {
                background-color: #cbd5e1;
                color: #1f2937;
            }
            .copy-icon, .check-icon {
                display: block;
            }
            .message-log {
                max-height: 120px;
                overflow-y: auto;
                border-top: 1px solid #edf2f7;
                padding: 3px 0;
                font-size: 0.85em;
            }
            .message-item {
                display: flex;
                padding: 2px 12px;
            }
            .message-timestamp {
                color: #9ca3af;
                font-size: 0.85em;
                margin-right: 8px;
                white-space: nowrap;
            }
            .status-indicator {
                display: inline-block;
                width: 8px;
                height: 8px;
                border-radius: 50%;
                margin-right: 8px;
            }
            .expand-toggle {
                display: inline-block;
                width: 16px;
                text-align: center;
                margin-right: 6px;
                font-size: 12px;
            }
            .badge {
                display: inline-block;
                padding: 2px 6px;
                border-radius: 3px;
                font-size: 0.8em;
                font-weight: 500;
            }
            .status-running.badge { background-color: #dbeafe; }
            .status-completed.badge { background-color: #d1fae5; }
            .status-partially-failed.badge { background-color: #fef3c7; }
            .status-failed.badge { background-color: #fee2e2; }
            .helper-text {
                color: #4b5563;
                font-size: 0.75em;
                text-align: left;
            }
            /* Exception table styles */
            .exception-section {
                border-top: 1px solid #edf2f7;
            }
            .exception-header {
                padding: 8px 12px;
                background-color: #f7fafc;
                display: flex;
                align-items: center;
                cursor: pointer;
                font-size: 0.85em;
                font-weight: 500;
                user-select: none;  /* Prevent text selection */
            }
            .exception-content {
                padding: 12px;
            }
            .exception-table {
                width: 100%;
                border-collapse: collapse;
                margin: 0;
                font-size: 0.85em;
            }
            .exception-table th {
                background-color: #f1f5f9;
                color: #475569;
                font-weight: 500;
                text-align: left;
                padding: 8px 12px;
                border-bottom: 2px solid #e2e8f0;
            }
            .exception-table td {
                padding: 6px 12px;
                border-bottom: 1px solid #e2e8f0;
                color: #1f2937;
                text-align: left;  /* Ensure left alignment */
            }
            .exception-table tr:last-child td {
                border-bottom: none;
            }
            .exception-count {
                font-weight: 500;
                color: #ef4444;
            }
            /* Model costs table styles */
            .model-costs-section {
                border-top: 1px solid #edf2f7;
            }
            .model-costs-header {
                padding: 8px 12px;
                background-color: #f7fafc;
                display: flex;
                align-items: center;
                cursor: pointer;
                font-size: 0.85em;
                font-weight: 500;
                user-select: none;
            }
            .model-costs-content {
                padding: 12px;
            }
            .model-costs-table {
                width: 100%;
                border-collapse: collapse;
                margin: 0;
                font-size: 0.85em;
            }
            .model-costs-table th {
                background-color: #f1f5f9;
                color: #475569;
                font-weight: 500;
                text-align: left;  /* Default left alignment */
                padding: 8px 12px;
                border-bottom: 2px solid #e2e8f0;
            }
            .model-costs-table th.cost-header {  /* New class for cost headers */
                text-align: right;
            }
            .model-costs-table td {
                padding: 6px 12px;
                border-bottom: 1px solid #e2e8f0;
                color: #1f2937;
                text-align: left;  /* Ensure left alignment for all cells by default */
            }
            .model-costs-table tr:last-child td {
                border-bottom: none;
            }
            .token-count td, .cost-value td {  /* Override for specific columns that need right alignment */
                text-align: right;
            }
            .totals-row {
                background-color: #f8fafc;
            }
            .totals-row td {
                border-top: 2px solid #e2e8f0;
            }
            /* Model costs table styles */
            .model-costs-table td.token-count,
            .model-costs-table td.cost-value {
                text-align: right;  /* Right align the token counts and cost values */
            }
            .code-text {
                font-family: "SF Mono", "Cascadia Mono", monospace;
                background-color: #f8fafc;
                padding: 1px 4px;
                border-radius: 3px;
            }
        </style>
        """

        # Group JobsInfo fields into categories
        url_fields = []
        uuid_fields = []
        other_fields = []

        for field, _ in self.jobs_info.__annotations__.items():
            if field not in [
                "pretty_names",
                "completed_interviews",
                "failed_interviews",
                "exception_summary",
                "model_costs",
            ]:
                value = getattr(self.jobs_info, field)
                if not value:
                    continue

                pretty_name = self.jobs_info.pretty_names.get(
                    field, field.replace("_", " ").title()
                )

                if "url" in field.lower():
                    url_fields.append((field, pretty_name, value))
                elif "uuid" in field.lower():
                    uuid_fields.append((field, pretty_name, value))
                else:
                    other_fields.append((field, pretty_name, value))

        # Build a three-column layout
        content_html = """
        <div class="three-column-grid">
            <div class="column">
                <div class="section-header">Job Links</div>
                <div class="content-box">
        """

        # Sort URLs to prioritize Results first, then Progress
        results_links = []
        progress_links = []
        remote_links = []
        other_links = []

        for field, pretty_name, value in url_fields:
            # Replace "Progress Bar" with "Progress Report"
            if "progress_bar" in field.lower():
                pretty_name = "Progress Report URL"

            label = pretty_name.replace(" URL", "")

            if "result" in field.lower():
                results_links.append((field, pretty_name, value, label))
            elif "progress" in field.lower() or "error_report" in field.lower():
                progress_links.append((field, pretty_name, value, label))
            elif "remote_cache" in field.lower() or "remote_inference" in field.lower():
                remote_links.append((field, pretty_name, value, label))
            else:
                other_links.append((field, pretty_name, value, label))

        # Add results and progress links to first column
        for field, pretty_name, value, label in results_links:
            content_html += f"""
            <div class="link-item results-link" style="display: flex; align-items: center; justify-content: space-between;">
                <a href="{value}" target="_blank" class="pill-link">{label}{self.external_link_icon}</a>
                {self._create_copy_button(value)}
            </div>
            """

        # Then add progress links with different special styling
        for field, pretty_name, value, label in progress_links:
            content_html += f"""
            <div class="link-item progress-link" style="display: flex; align-items: center; justify-content: space-between;">
                <a href="{value}" target="_blank" class="pill-link">{label}{self.external_link_icon}</a>
                {self._create_copy_button(value)}
            </div>
            """

        # Close first column and start second column
        content_html += """
                </div>
            </div>
            <div class="column">
                <div class="section-header">Content</div>
                <div class="content-box">
        """

        # Add remote links to middle column
        for field, pretty_name, value, label in remote_links + other_links:
            content_html += f"""
            <div class="link-item remote-link" style="display: flex; align-items: center; justify-content: space-between;">
                <a href="{value}" target="_blank" class="pill-link">{label}{self.external_link_icon}</a>
                {self._create_copy_button(value)}
            </div>
            """

        content_html += """
                </div>
            </div>
            <div class="column">
                <div class="section-header">Identifiers</div>
                <div class="content-box">
        """

        # Sort UUIDs to prioritize Result UUID first
        uuid_fields.sort(key=lambda x: 0 if "result" in x[0].lower() else 1)
        for field, pretty_name, value in uuid_fields:
            if "result" in field.lower():
                helper_text = "Use <span class='code-text'>Results.pull(uuid)</span> to fetch results."
            elif "job" in field.lower():
                helper_text = (
                    "Use <span class='code-text'>Jobs.pull(uuid)</span> to fetch job."
                )
            else:
                helper_text = ""

            content_html += f"""
            <div class="uuid-item">
                <span class="uuid-label">{pretty_name}:</span>{self._create_uuid_copy_button(value, helper_text)}
            </div>
            """

        content_html += """
                </div>
            </div>
        </div>
        """

        # Add other fields in full width if any
        if other_fields:
            content_html += "<div class='section-header'>Additional Information</div>"
            content_html += "<table class='jobs-table'>"
            for field, pretty_name, value in other_fields:
                content_html += f"""
                <tr>
                    <th>{pretty_name}</th>
                    <td>{value}</td>
                </tr>
                """
            content_html += "</table>"

        # Status banner and message log
        status_class = {
            JobsStatus.RUNNING: "status-running",
            JobsStatus.COMPLETED: "status-completed",
            JobsStatus.PARTIALLY_FAILED: "status-partially-failed",
            JobsStatus.FAILED: "status-failed",
        }.get(current_status, "status-unknown")

        status_icon = self._get_status_icon(current_status)
        status_text = current_status.name.capitalize()
        if current_status == JobsStatus.PARTIALLY_FAILED:
            status_text = "Partially failed"
        elif hasattr(current_status, "name"):
            status_text = current_status.name.capitalize()
        else:
            status_text = str(current_status).capitalize()

        status_banner = f"""
        <div class="status-banner" onclick="{self._collapse(f'message-log-{self.log_id}', f'message-arrow-{self.log_id}')}">
            <div style="display: flex; align-items: center; gap: 8px;">
                <span id="message-arrow-{self.log_id}" class="expand-toggle">&#8963;</span>
                <div style="display: flex; align-items: center;">
                    {status_icon}
                    <strong>Status:</strong>&nbsp;<span class="badge {status_class}">{status_text}</span>
                </div>
            </div>
            <span style="flex-grow: 1;"></span>
            <span>Last updated: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}</span>
        </div>
        """

        # Message history
        message_log = ""
        if self.messages:
            message_items = []
            for msg in self.messages:
                status_color = {
                    JobsStatus.RUNNING: "#3b82f6",
                    JobsStatus.COMPLETED: "#10b981",
                    JobsStatus.PARTIALLY_FAILED: "#f59e0b",
                    JobsStatus.FAILED: "#ef4444",
                }.get(msg["status"], "#6b7280")

                message_items.append(
                    f"""
                <div class="message-item">
                    <span class="message-timestamp">{msg["timestamp"]}</span>
                    <span class="status-indicator" style="background-color: {status_color};"></span>
                    <div>{self._linkify(msg["text"])}</div>
                </div>
                """
                )

            message_log = f"""
            <div id="message-log-{self.log_id}" class="message-log">
                {''.join(reversed(message_items))}
            </div>
            """

        display_style = "block" if self.is_expanded else "none"

        header_status_text = status_text
        if (
            self.jobs_info.completed_interviews is not None
            and self.jobs_info.failed_interviews is not None
        ):
            header_status_text += f" ({self.jobs_info.completed_interviews:,} completed, {self.jobs_info.failed_interviews:,} failed)"

        # Add model costs table before exceptions table
        main_content = f"""
            {content_html}
            {status_banner}
            {message_log}
            {self._build_model_costs_table()}
            {self._build_exceptions_table()}
        """

        # Return the complete HTML
        return f"""
        {css}
        <div class="jobs-container">
            <div class="jobs-header" onclick="{self._collapse(f'content-{self.log_id}', f'arrow-{self.log_id}')}">
                <div>
                    <span id="arrow-{self.log_id}" class="expand-toggle">{'&#8963;' if self.is_expanded else '&#8964;'}</span>
                    Job Status 🦜
                </div>
                <div class="{status_class}">{header_status_text}</div>
            </div>
            <div id="content-{self.log_id}" class="jobs-content" style="display: {display_style};">
                {main_content}
            </div>
        </div>
        """



"""
Memory debugging utilities for analyzing object references and memory leaks.
"""
import gc
import os
import sys
import time
import types
import inspect
import webbrowser
from pathlib import Path
import base64
import html
from datetime import datetime
from typing import Any, Set, List, Dict, Optional, Tuple, Union

# Try to import objgraph, which is only available as a dev dependency
try:
    import objgraph
    OBJGRAPH_AVAILABLE = True
except ImportError:
    OBJGRAPH_AVAILABLE = False

class MemoryDebugger:
    """
    A class for debugging memory issues and analyzing object references.
    
    This class provides utilities to:
    - Inspect objects referring to a target object
    - Detect reference cycles
    - Visualize object dependencies
    - Analyze memory usage patterns
    """
    
    def __init__(self, target_obj: Any):
        """
        Initialize the debugger with a target object to analyze.
        
        Args:
            target_obj: The object to inspect for memory issues
        """
        self.target_obj = target_obj
    
    def _get_source_info(self, obj: Any) -> Dict[str, Any]:
        """Get source code information for an object if available."""
        result = {
            'module': getattr(obj, '__module__', 'unknown'),
            'file': 'unknown',
            'line': 0,
            'source': None,
            'has_source': False,
            'type_str': str(type(obj))
        }
        
        try:
            # Try to get the source file and line number
            if hasattr(obj, '__class__'):
                module = inspect.getmodule(obj.__class__)
                if module:
                    result['module'] = module.__name__
                    try:
                        file = inspect.getsourcefile(obj.__class__)
                        if file:
                            result['file'] = file
                            try:
                                _, line = inspect.getsourcelines(obj.__class__)
                                result['line'] = line
                                result['has_source'] = True
                            except Exception:
                                pass
                    except Exception:
                        pass
        except Exception:
            # Silently fail if we can't get source info
            pass
            
        return result
        
    def _generate_html_report(self, prefix: str = "", output_dir: str = None) -> Tuple[str, str, str]:
        """
        Generate a comprehensive HTML memory debugging report.
        
        Args:
            prefix: Optional prefix for output files. If empty, uses target object type.
            output_dir: Optional directory to write files to. If None, uses tempdir or current directory.
            
        Returns:
            A tuple containing (html_filename, refs_graph_filename, backrefs_graph_filename)
        """
        timestamp = time.strftime("%Y%m%d_%H%M%S")
        if not prefix:
            prefix = type(self.target_obj).__name__.lower()
            
        # Determine output directory
        if output_dir is None:
            output_dir = os.environ.get("EDSL_MEMORY_DEBUG_DIR", "")
            
        if output_dir:
            # Ensure directory exists
            os.makedirs(output_dir, exist_ok=True)
            
        # Prepare filenames
        html_filename = os.path.join(output_dir, f"{prefix}_memory_debug_{timestamp}.html") if output_dir else f"{prefix}_memory_debug_{timestamp}.html"
        refs_graph_filename = os.path.join(output_dir, f"{prefix}_outgoing_refs_{timestamp}.png") if output_dir else f"{prefix}_outgoing_refs_{timestamp}.png"
        backrefs_graph_filename = os.path.join(output_dir, f"{prefix}_incoming_refs_{timestamp}.png") if output_dir else f"{prefix}_incoming_refs_{timestamp}.png"
        
        # Generate object graphs - both directions
        graph_exists = False
        backrefs_graph_exists = False
        
        if OBJGRAPH_AVAILABLE:
            try:
                # Use objgraph's backup function to filter frames and functions
                def skip_frames(obj):
                    return not isinstance(obj, (types.FrameType, types.FunctionType))
                
                # Outgoing references (what this object references)
                objgraph.show_refs(
                    self.target_obj, 
                    filename=refs_graph_filename,
                    max_depth=3,
                    extra_ignore=[id(obj) for obj in gc.get_objects() 
                                if isinstance(obj, (types.FrameType, types.FunctionType))]
                )
                graph_exists = True
                
                # Incoming references (what references this object)
                objgraph.show_backrefs(
                    self.target_obj, 
                    filename=backrefs_graph_filename,
                    max_depth=3,
                    extra_ignore=[id(obj) for obj in gc.get_objects() 
                                if isinstance(obj, (types.FrameType, types.FunctionType))]
                )
                backrefs_graph_exists = True
            except Exception as e:
                print(f"Warning: Could not generate object graph visualization: {e}")
                graph_exists = False
        else:
            print("Warning: objgraph package is not available. Install it with 'pip install objgraph' to enable visualizations.")
        
        # Get all reference cycle information
        import io
        from contextlib import redirect_stdout
        
        captured_output = io.StringIO()
        with redirect_stdout(captured_output):
            cycles = self.detect_reference_cycles()
            
        cycle_output = captured_output.getvalue()
        
        # Get referrers and referents with source info
        referrers = gc.get_referrers(self.target_obj)
        referents = gc.get_referents(self.target_obj)
        
        # Skip frames and function objects for referrers
        filtered_referrers = [ref for ref in referrers if not isinstance(ref, (types.FrameType, types.FunctionType))]
        
        # Get source info for all objects
        referrer_info = [self._get_source_info(ref) for ref in filtered_referrers]
        referent_info = [self._get_source_info(ref) for ref in referents]
        
        # Get target object info
        target_info = self._get_source_info(self.target_obj)
        
        # Generate HTML content
        html_content = f"""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Memory Debug Report - {type(self.target_obj).__name__}</title>
    <style>
        body {{
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, Cantarell, 'Open Sans', 'Helvetica Neue', sans-serif;
            line-height: 1.6;
            color: #333;
            max-width: 1200px;
            margin: 0 auto;
            padding: 20px;
        }}
        h1, h2, h3 {{
            color: #2c3e50;
        }}
        h1 {{
            border-bottom: 2px solid #eaecef;
            padding-bottom: 10px;
        }}
        h2 {{
            margin-top: 30px;
            border-bottom: 1px solid #eaecef;
            padding-bottom: 5px;
        }}
        .info-box {{
            background-color: #f8f9fa;
            border: 1px solid #e9ecef;
            border-radius: 5px;
            padding: 15px;
            margin: 20px 0;
        }}
        .warning {{
            background-color: #fff3cd;
            border-color: #ffeeba;
        }}
        .object {{
            margin-bottom: 10px;
            padding: 10px;
            background-color: #f8f9fa;
            border-left: 4px solid #007bff;
            border-radius: 3px;
        }}
        .object-header {{
            display: flex;
            justify-content: space-between;
            font-weight: bold;
        }}
        .object-details {{
            margin-top: 5px;
            font-size: 0.9em;
        }}
        .file-link {{
            color: #007bff;
            text-decoration: none;
        }}
        .file-link:hover {{
            text-decoration: underline;
        }}
        .unhashable {{
            border-left-color: #dc3545;
        }}
        .cycle {{
            border-left-color: #fd7e14;
        }}
        .tab {{
            overflow: hidden;
            border: 1px solid #ccc;
            background-color: #f1f1f1;
            border-radius: 5px 5px 0 0;
        }}
        .tab button {{
            background-color: inherit;
            float: left;
            border: none;
            outline: none;
            cursor: pointer;
            padding: 10px 16px;
            transition: 0.3s;
        }}
        .tab button:hover {{
            background-color: #ddd;
        }}
        .tab button.active {{
            background-color: #ccc;
        }}
        .tabcontent {{
            display: none;
            padding: 20px;
            border: 1px solid #ccc;
            border-top: none;
            border-radius: 0 0 5px 5px;
        }}
        .code {{
            font-family: 'SFMono-Regular', Consolas, 'Liberation Mono', Menlo, monospace;
            background-color: #f6f8fa;
            padding: 10px;
            border-radius: 3px;
            overflow-x: auto;
        }}
        .graph-container {{
            text-align: center;
            margin: 20px 0;
        }}
        table {{
            border-collapse: collapse;
            width: 100%;
            margin: 20px 0;
        }}
        th, td {{
            border: 1px solid #ddd;
            padding: 8px;
            text-align: left;
        }}
        th {{
            background-color: #f2f2f2;
        }}
        tr:nth-child(even) {{
            background-color: #f9f9f9;
        }}
        .toggle-button {{
            background-color: #4CAF50;
            border: none;
            color: white;
            padding: 5px 10px;
            text-align: center;
            text-decoration: none;
            display: inline-block;
            font-size: 12px;
            margin: 4px 2px;
            cursor: pointer;
            border-radius: 3px;
        }}
    </style>
</head>
<body>
    <h1>Memory Debug Report - {type(self.target_obj).__name__}</h1>
    <p>Generated: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}</p>
    
    <div class="info-box">
        <h3>Target Object Information</h3>
        <p><strong>Type:</strong> {type(self.target_obj)}</p>
        <p><strong>ID:</strong> {id(self.target_obj)}</p>
        <p><strong>Module:</strong> {target_info['module']}</p>
        <p><strong>Reference Count:</strong> {sys.getrefcount(self.target_obj)}</p>
        {f'<p><strong>Source File:</strong> <a href="file://{target_info["file"]}" class="file-link">{os.path.basename(target_info["file"])}</a> (line {target_info["line"]})</p>' if target_info["has_source"] else ''}
    </div>
    
    <div class="tab">
        <button class="tablinks" onclick="openTab(event, 'Overview')" id="defaultOpen">Overview</button>
        <button class="tablinks" onclick="openTab(event, 'Referrers')">Referrers ({len(filtered_referrers)})</button>
        <button class="tablinks" onclick="openTab(event, 'Referents')">Referents ({len(referents)})</button>
        <button class="tablinks" onclick="openTab(event, 'Cycles')" {'style="color: red;"' if cycles else ''}>Cycles {f"({len(cycles)})" if cycles else ""}</button>
        <button class="tablinks" onclick="openTab(event, 'Graph')">Reference Visualizations</button>
    </div>
    
    <div id="Overview" class="tabcontent">
        <h2>Memory Analysis Overview</h2>
        <div class="info-box">
            <p>This report analyzes the memory references to and from the target object. It helps identify potential memory leaks by showing object relationships.</p>
            <p><strong>Total referrers:</strong> {len(filtered_referrers)}</p>
            <p><strong>Total referents:</strong> {len(referents)}</p>
            <p><strong>Potential reference cycles:</strong> {len(cycles) if cycles else "None detected"}</p>
        </div>
        
        <h3>Common Types</h3>
        <table>
            <tr>
                <th>Type</th>
                <th>Count as Referrer</th>
                <th>Count as Referent</th>
            </tr>
"""

        # Count types
        referrer_types = {}
        for ref in filtered_referrers:
            type_name = type(ref).__name__
            referrer_types[type_name] = referrer_types.get(type_name, 0) + 1
            
        referent_types = {}
        for ref in referents:
            type_name = type(ref).__name__
            referent_types[type_name] = referent_types.get(type_name, 0) + 1
            
        # Get all unique types
        all_types = set(list(referrer_types.keys()) + list(referent_types.keys()))
        
        # Add type rows to the table
        for type_name in sorted(all_types):
            html_content += f"""
            <tr>
                <td>{type_name}</td>
                <td>{referrer_types.get(type_name, 0)}</td>
                <td>{referent_types.get(type_name, 0)}</td>
            </tr>"""
        
        html_content += """
        </table>
    </div>
    
    <div id="Referrers" class="tabcontent">
        <h2>Objects Referring to Target</h2>
        <p>These objects hold references to the target object:</p>
"""

        # Add referrers
        for i, (ref, info) in enumerate(zip(filtered_referrers, referrer_info)):
            object_id = id(ref)
            object_type = type(ref).__name__
            is_unhashable = not self._is_hashable(ref)
            
            html_content += f"""
        <div class="object {'unhashable' if is_unhashable else ''}">
            <div class="object-header">
                <span>[{i}] {object_type}{' (unhashable)' if is_unhashable else ''}</span>
                <span>ID: {object_id}</span>
            </div>
            <div class="object-details">
                <p><strong>Module:</strong> {info['module']}</p>
                {f'<p><strong>Source:</strong> <a href="file://{info["file"]}" class="file-link">{os.path.basename(info["file"])}</a> (line {info["line"]})</p>' if info["has_source"] else ''}"""
            
            # Add more specific information based on type
            if isinstance(ref, dict):
                html_content += f"""
                <p><strong>Dict size:</strong> {len(ref)} items</p>
                <button class="toggle-button" onclick="toggleDetails('referrer-{i}')">Show/Hide Contents</button>
                <div id="referrer-{i}" style="display: none;">
                    <table>
                        <tr><th>Key</th><th>Value Type</th></tr>"""
                
                for k, v in list(ref.items())[:20]:  # Limit to 20 items
                    # Skip showing the target object itself
                    if v is self.target_obj:
                        key_html = html.escape(str(k))
                        html_content += f"""
                        <tr><td>{key_html}</td><td><strong>TARGET OBJECT</strong></td></tr>"""
                    else:
                        key_html = html.escape(str(k))
                        html_content += f"""
                        <tr><td>{key_html}</td><td>{type(v).__name__}</td></tr>"""
                
                if len(ref) > 20:
                    html_content += f"""
                        <tr><td colspan="2">... and {len(ref) - 20} more items</td></tr>"""
                    
                html_content += """
                    </table>
                </div>"""
            elif isinstance(ref, (list, tuple)):
                html_content += f"""
                <p><strong>{type(ref).__name__} size:</strong> {len(ref)} items</p>
                <button class="toggle-button" onclick="toggleDetails('referrer-{i}')">Show/Hide Contents</button>
                <div id="referrer-{i}" style="display: none;">
                    <table>
                        <tr><th>Index</th><th>Value Type</th></tr>"""
                
                for idx, item in enumerate(ref[:20]):  # Limit to 20 items
                    if item is self.target_obj:
                        html_content += f"""
                        <tr><td>{idx}</td><td><strong>TARGET OBJECT</strong></td></tr>"""
                    else:
                        html_content += f"""
                        <tr><td>{idx}</td><td>{type(item).__name__}</td></tr>"""
                
                if len(ref) > 20:
                    html_content += f"""
                        <tr><td colspan="2">... and {len(ref) - 20} more items</td></tr>"""
                    
                html_content += """
                    </table>
                </div>"""
            
            html_content += """
            </div>
        </div>"""
        
        html_content += """
    </div>
    
    <div id="Referents" class="tabcontent">
        <h2>Objects Referenced by Target</h2>
        <p>These objects are referenced by the target object:</p>
"""

        # Add referents
        for i, (ref, info) in enumerate(zip(referents, referent_info)):
            object_id = id(ref)
            object_type = type(ref).__name__
            is_unhashable = not self._is_hashable(ref)
            
            html_content += f"""
        <div class="object {'unhashable' if is_unhashable else ''}">
            <div class="object-header">
                <span>[{i}] {object_type}{' (unhashable)' if is_unhashable else ''}</span>
                <span>ID: {object_id}</span>
            </div>
            <div class="object-details">
                <p><strong>Module:</strong> {info['module']}</p>
                {f'<p><strong>Source:</strong> <a href="file://{info["file"]}" class="file-link">{os.path.basename(info["file"])}</a> (line {info["line"]})</p>' if info["has_source"] else ''}"""
            
            # Add specific information for dicts and sequences
            if isinstance(ref, dict):
                html_content += f"""
                <p><strong>Dict size:</strong> {len(ref)} items</p>
                <button class="toggle-button" onclick="toggleDetails('referent-{i}')">Show/Hide Contents</button>
                <div id="referent-{i}" style="display: none;">
                    <table>
                        <tr><th>Key</th><th>Value Type</th></tr>"""
                
                for k, v in list(ref.items())[:20]:  # Limit to 20 items
                    key_html = html.escape(str(k))
                    html_content += f"""
                        <tr><td>{key_html}</td><td>{type(v).__name__}</td></tr>"""
                
                if len(ref) > 20:
                    html_content += f"""
                        <tr><td colspan="2">... and {len(ref) - 20} more items</td></tr>"""
                    
                html_content += """
                    </table>
                </div>"""
            elif isinstance(ref, (list, tuple)):
                html_content += f"""
                <p><strong>{type(ref).__name__} size:</strong> {len(ref)} items</p>
                <button class="toggle-button" onclick="toggleDetails('referent-{i}')">Show/Hide Contents</button>
                <div id="referent-{i}" style="display: none;">
                    <table>
                        <tr><th>Index</th><th>Value Type</th></tr>"""
                
                for idx, item in enumerate(ref[:20]):  # Limit to 20 items
                    html_content += f"""
                        <tr><td>{idx}</td><td>{type(item).__name__}</td></tr>"""
                
                if len(ref) > 20:
                    html_content += f"""
                        <tr><td colspan="2">... and {len(ref) - 20} more items</td></tr>"""
                    
                html_content += """
                    </table>
                </div>"""
            
            html_content += """
            </div>
        </div>"""
        
        html_content += """
    </div>
    
    <div id="Cycles" class="tabcontent">
        <h2>Reference Cycles</h2>
"""

        # Add reference cycle output
        if cycles:
            html_content += f"""
        <div class="info-box warning">
            <p>Found {len(cycles)} potential reference cycles. These objects might be causing memory leaks.</p>
        </div>
"""
            # Add cycle details
            for i, obj in enumerate(cycles):
                object_type = type(obj).__name__
                object_id = id(obj)
                info = self._get_source_info(obj)
                
                html_content += f"""
        <div class="object cycle">
            <div class="object-header">
                <span>[{i}] {object_type}</span>
                <span>ID: {object_id}</span>
            </div>
            <div class="object-details">
                <p><strong>Module:</strong> {info['module']}</p>
                {f'<p><strong>Source:</strong> <a href="file://{info["file"]}" class="file-link">{os.path.basename(info["file"])}</a> (line {info["line"]})</p>' if info["has_source"] else ''}
            </div>
        </div>"""
        else:
            html_content += """
        <p>No reference cycles detected among hashable objects.</p>
"""
        
        # Add unhashable object analysis
        html_content += """
        <h3>Unhashable Object Analysis</h3>
        <div class="code">
"""
        for line in cycle_output.splitlines():
            html_content += f"{html.escape(line)}<br>"
            
        html_content += """
        </div>
    </div>
    
    <div id="Graph" class="tabcontent">
        <h2>Object Graph Visualization</h2>
        
        <h3>Outgoing References (Objects Referenced By Target)</h3>
        <p>This graph shows what objects are referenced by the target object:</p>
"""

        if graph_exists:
            try:
                with open(refs_graph_filename, 'rb') as f:
                    img_data = base64.b64encode(f.read()).decode('utf-8')
                html_content += f"""
        <div class="graph-container">
            <img src="data:image/png;base64,{img_data}" alt="Objects referenced by target" style="max-width: 100%;">
        </div>
        <p><a href="file://{os.path.abspath(refs_graph_filename)}" target="_blank">Open full-size outgoing references graph</a></p>
"""
            except Exception as e:
                html_content += f"""
        <div class="info-box warning">
            <p>Could not embed outgoing references graph image: {e}</p>
            <p><a href="file://{os.path.abspath(refs_graph_filename)}" target="_blank">Open outgoing references graph image</a></p>
        </div>
"""
        else:
            html_content += """
        <div class="info-box warning">
            <p>Could not generate outgoing references graph visualization.</p>
        </div>
"""

        html_content += """
        <h3>Incoming References (Objects Referencing Target)</h3>
        <p>This graph shows what objects have a strong reference to the target object:</p>
"""

        if backrefs_graph_exists:
            try:
                with open(backrefs_graph_filename, 'rb') as f:
                    img_data = base64.b64encode(f.read()).decode('utf-8')
                html_content += f"""
        <div class="graph-container">
            <img src="data:image/png;base64,{img_data}" alt="Objects referencing the target" style="max-width: 100%;">
        </div>
        <p><a href="file://{os.path.abspath(backrefs_graph_filename)}" target="_blank">Open full-size incoming references graph</a></p>
"""
            except Exception as e:
                html_content += f"""
        <div class="info-box warning">
            <p>Could not embed incoming references graph image: {e}</p>
            <p><a href="file://{os.path.abspath(backrefs_graph_filename)}" target="_blank">Open incoming references graph image</a></p>
        </div>
"""
        else:
            html_content += """
        <div class="info-box warning">
            <p>Could not generate incoming references graph visualization.</p>
        </div>
"""
        
        html_content += """
    </div>
    
    <script>
        function openTab(evt, tabName) {
            var i, tabcontent, tablinks;
            tabcontent = document.getElementsByClassName("tabcontent");
            for (i = 0; i < tabcontent.length; i++) {
                tabcontent[i].style.display = "none";
            }
            tablinks = document.getElementsByClassName("tablinks");
            for (i = 0; i < tablinks.length; i++) {
                tablinks[i].className = tablinks[i].className.replace(" active", "");
            }
            document.getElementById(tabName).style.display = "block";
            evt.currentTarget.className += " active";
        }
        
        function toggleDetails(id) {
            var element = document.getElementById(id);
            if (element.style.display === "none") {
                element.style.display = "block";
            } else {
                element.style.display = "none";
            }
        }
        
        // Open the default tab
        document.getElementById("defaultOpen").click();
    </script>
</body>
</html>
"""

        # Write HTML to file
        with open(html_filename, 'w') as f:
            f.write(html_content)
            
        return html_filename, refs_graph_filename, backrefs_graph_filename
        
    def debug_memory(self, prefix: str = "", open_browser: bool = True, output_dir: str = None) -> str:
        """
        Comprehensive memory debugging that writes results to files and opens an HTML report.
        
        Args:
            prefix: Optional prefix for output files. If empty, uses target object type.
            open_browser: Whether to automatically open the HTML report in a browser.
            output_dir: Optional directory to write files to. If None, uses EDSL_MEMORY_DEBUG_DIR
                        environment variable or current directory.
        
        Returns:
            The path to the HTML report file.
        """
        # Generate HTML report with both incoming and outgoing reference visualizations
        html_filename, outgoing_refs_filename, incoming_refs_filename = self._generate_html_report(prefix, output_dir)
        
        # Also create the legacy markdown report for backward compatibility
        timestamp = time.strftime("%Y%m%d_%H%M%S")
        if not prefix:
            prefix = type(self.target_obj).__name__.lower()
        
        # Determine output directory
        if output_dir is None:
            output_dir = os.environ.get("EDSL_MEMORY_DEBUG_DIR", "")
            
        # Write reference analysis to markdown file
        md_filename = os.path.join(output_dir, f"{prefix}_memory_debug_{timestamp}.md") if output_dir else f"{prefix}_memory_debug_{timestamp}.md"
        with open(md_filename, 'w') as f:
            f.write(f"# Memory Debug Report for {type(self.target_obj)}\n\n")
            f.write(f"Generated: {time.strftime('%Y-%m-%d %H:%M:%S')}\n\n")
            f.write(f"HTML Report: [View HTML Report]({html_filename})\n\n")
            f.write(f"Visualizations:\n")
            f.write(f"- [Outgoing References Graph]({outgoing_refs_filename}) - Objects referenced by the target\n")
            f.write(f"- [Incoming References Graph]({incoming_refs_filename}) - Objects that reference the target\n\n")
            
            # Capture reference count
            f.write(f"## Reference Count\n")
            f.write(f"Current reference count: {sys.getrefcount(self.target_obj)}\n\n")
            
            # Capture referrers
            f.write(f"## Objects Referring to Target\n")
            referrers = gc.get_referrers(self.target_obj)
            for ref in referrers:
                if isinstance(ref, (types.FrameType, types.FunctionType)):
                    continue
                f.write(f"- Type: {type(ref)}\n")
            
            # Capture reference cycles
            f.write(f"\n## Reference Cycles\n")
            
            # Get all reference cycle information (capture stdout for the detailed info)
            import io
            from contextlib import redirect_stdout
            
            captured_output = io.StringIO()
            with redirect_stdout(captured_output):
                cycles = self.detect_reference_cycles()
            
            # Write the cycle information to file
            cycle_output = captured_output.getvalue()
            if cycles:
                f.write(f"Found {len(cycles)} potential reference cycles:\n")
                for obj in cycles:
                    f.write(f"- Type: {type(obj)}, ID: {id(obj)}\n")
            else:
                f.write("No reference cycles detected among hashable objects\n")
                
            # Add the detailed unhashable information
            f.write("\n## Unhashable Object Analysis\n")
            f.write(cycle_output)
        
        # Open the HTML report in a browser if requested
        if open_browser:
            try:
                webbrowser.open(f"file://{os.path.abspath(html_filename)}")
                print(f"Opened memory report in browser: {html_filename}")
            except Exception as e:
                print(f"Could not open browser: {e}")
                print(f"Report saved to: {html_filename}")
        else:
            print(f"HTML Report saved to: {html_filename}")
            
        return html_filename
    
    def inspect_references(self, skip_frames: bool = True) -> None:
        """
        Inspect what objects are referring to the target object.
        
        Args:
            skip_frames: If True, skip function frames and local namespaces
        """
        print(f"\nReference count for {type(self.target_obj)}: {sys.getrefcount(self.target_obj)}")
        print("\nObjects referring to this object:")
        
        referrers = gc.get_referrers(self.target_obj)
        for ref in referrers:
            # Skip frames and function locals if requested
            if skip_frames and (isinstance(ref, (types.FrameType, types.FunctionType)) or 
                              (isinstance(ref, dict) and ref.get('target_obj') is self.target_obj)):
                continue
        
            print(f"\nType: {type(ref)}")
            
            if isinstance(ref, dict):
                self._inspect_dict_reference(ref)
            elif isinstance(ref, list):
                self._inspect_list_reference(ref)
            elif isinstance(ref, tuple):
                self._inspect_tuple_reference(ref)
            else:
                print(f"  - {ref}")
    
    def detect_reference_cycles(self) -> Set[Any]:
        """
        Detect potential reference cycles involving the target object.
        
        Returns:
            Set of objects that are part of potential reference cycles
        """
        referrers = gc.get_referrers(self.target_obj)
        referents = gc.get_referents(self.target_obj)
        
        # Separate hashable and unhashable objects
        hashable_referrers = []
        unhashable_referrers = []
        hashable_referents = []
        unhashable_referents = []
        
        for obj in referrers:
            if self._is_hashable(obj):
                hashable_referrers.append(obj)
            else:
                unhashable_referrers.append(obj)
                
        for obj in referents:
            if self._is_hashable(obj):
                hashable_referents.append(obj)
            else:
                unhashable_referents.append(obj)
        
        # Find common objects among hashable ones
        common_objects = set(hashable_referrers) & set(hashable_referents)
        
        if common_objects:
            print(f"Potential reference cycle detected! Found {len(common_objects)} common objects")
            for shared_obj in common_objects:
                print(f"Type: {type(shared_obj)}, ID: {id(shared_obj)}")
        
        # Report on unhashable objects
        if unhashable_referrers or unhashable_referents:
            print(f"Note: {len(unhashable_referrers)} unhashable referrers and {len(unhashable_referents)} unhashable referents were excluded from cycle detection")
            
            # Check for unhashable objects that might be in both lists
            potential_unhashable_cycles = []
            for ureferrer in unhashable_referrers:
                for ureferent in unhashable_referents:
                    if id(ureferrer) == id(ureferent):
                        potential_unhashable_cycles.append(ureferrer)
            
            if potential_unhashable_cycles:
                print(f"Warning: Found {len(potential_unhashable_cycles)} unhashable objects that may be part of cycles:")
                for obj in potential_unhashable_cycles:
                    print(f"  - Type: {type(obj)}, ID: {id(obj)}")
            
            # Report details of unhashable objects for further investigation
            print("Unhashable referrers:")
            for i, obj in enumerate(unhashable_referrers[:5]):  # Limit to first 5 to avoid flooding output
                print(f"  - [{i}] Type: {type(obj)}, ID: {id(obj)}")
            if len(unhashable_referrers) > 5:
                print(f"  ... and {len(unhashable_referrers) - 5} more")
                
            print("Unhashable referents:")
            for i, obj in enumerate(unhashable_referents[:5]):  # Limit to first 5
                print(f"  - [{i}] Type: {type(obj)}, ID: {id(obj)}")
            if len(unhashable_referents) > 5:
                print(f"  ... and {len(unhashable_referents) - 5} more")
        
        # Add unhashable objects that might be in cycles to the full result list
        result_set = common_objects.copy()
        return result_set
        
    def _is_hashable(self, obj: Any) -> bool:
        """Helper method to check if an object is hashable."""
        try:
            hash(obj)
            return True
        except TypeError:
            return False
    
    def visualize_dependencies(self, max_depth: int = 3, output_dir: str = None, prefix: str = "") -> Tuple[str, str]:
        """
        Visualize object dependencies using objgraph.
        
        Args:
            max_depth: Maximum depth of reference tree to display
            output_dir: Directory to save visualization files
            prefix: Prefix for output files
            
        Returns:
            Tuple of (outgoing_refs_filename, incoming_refs_filename)
        """
        if not OBJGRAPH_AVAILABLE:
            print("Warning: objgraph package is not available. Install it with 'pip install objgraph' to enable visualizations.")
            return "", ""
            
        timestamp = time.strftime("%Y%m%d_%H%M%S")
        if not prefix:
            prefix = type(self.target_obj).__name__.lower()
            
        # Determine output directory
        if output_dir is None:
            output_dir = os.environ.get("EDSL_MEMORY_DEBUG_DIR", "")
            
        if output_dir:
            # Ensure directory exists
            os.makedirs(output_dir, exist_ok=True)
        
        # Prepare filenames  
        refs_filename = os.path.join(output_dir, f"{prefix}_outgoing_refs_{timestamp}.png") if output_dir else f"{prefix}_outgoing_refs_{timestamp}.png"
        backrefs_filename = os.path.join(output_dir, f"{prefix}_incoming_refs_{timestamp}.png") if output_dir else f"{prefix}_incoming_refs_{timestamp}.png"
        
        try:
            # Filter out frames and functions
            ignore_ids = [id(obj) for obj in gc.get_objects() 
                          if isinstance(obj, (types.FrameType, types.FunctionType))]
                          
            # For the regular references (what the object references)
            objgraph.show_refs(
                self.target_obj, 
                max_depth=max_depth, 
                filename=refs_filename,
                extra_ignore=ignore_ids
            )
            
            # For the referrers (what references the object)
            objgraph.show_backrefs(
                self.target_obj, 
                max_depth=max_depth, 
                filename=backrefs_filename,
                extra_ignore=ignore_ids
            )
            
            print(f"Outgoing references saved to: {refs_filename}")
            print(f"Incoming references saved to: {backrefs_filename}")
            
            return refs_filename, backrefs_filename
        except Exception as e:
            print(f"Error visualizing dependencies: {e}")
            return "", ""
    
    def _inspect_dict_reference(self, ref: Dict) -> None:
        """Helper method to inspect dictionary references."""
        for k, v in ref.items():
            if v is self.target_obj:
                print(f"  - Found in dict with key: {k}")
                try:
                    owner = [o for o in gc.get_referrers(ref) 
                            if hasattr(o, '__dict__') and o.__dict__ is ref]
                    if owner:
                        print(f"    (This dict belongs to: {type(owner[0])})")
                except:
                    pass
    
    def _inspect_list_reference(self, ref: List) -> None:
        """Helper method to inspect list references."""
        try:
            idx = ref.index(self.target_obj)
            print(f"  - Found in list at index: {idx}")
            owners = [o for o in gc.get_referrers(ref) if hasattr(o, '__dict__')]
            if owners:
                print(f"    (This list belongs to: {type(owners[0])})")
        except ValueError:
            print("  - Found in list (as part of a larger structure)")
    
    def _inspect_tuple_reference(self, ref: tuple) -> None:
        """Helper method to inspect tuple references."""
        try:
            idx = ref.index(self.target_obj)
            print(f"  - Found in tuple at index: {idx}")
        except ValueError:
            print("  - Found in tuple (as part of a larger structure)")
            
    def find_reference_paths(self, max_depth: int = 5) -> None:
        """
        Find and print paths to objects that reference the target object.
        This is a simplified version that directly prints the results.
        
        Args:
            max_depth: Maximum depth to search for references
        """
        print(f"\nFinding reference paths to {type(self.target_obj)} (id: {id(self.target_obj)}):")
        
        # Create a simplified reference chain explorer
        def find_path_to_referrers(obj, path=None, depth=0, visited=None):
            if path is None:
                path = []
            if visited is None:
                visited = set()
                
            if depth > max_depth:
                return
                
            # Get referrers excluding frames and functions
            referrers = [ref for ref in gc.get_referrers(obj) 
                        if not isinstance(ref, (types.FrameType, types.FunctionType))]
            
            for ref in referrers:
                ref_id = id(ref)
                
                # Skip if we've seen this object already
                if ref_id in visited:
                    continue
                    
                visited.add(ref_id)
                
                # Print current path
                ref_type = type(ref).__name__
                current_path = path + [(ref_type, ref_id)]
                
                # Print the path
                path_str = " -> ".join([f"{t} (id:{i})" for t, i in current_path])
                print(f"{' ' * depth}• {ref_type} references {type(obj).__name__}")
                
                # If it's a container, try to find the specific reference
                if isinstance(ref, dict):
                    for k, v in ref.items():
                        if v is obj:
                            print(f"{' ' * (depth+2)}(via dict key: {k})")
                elif isinstance(ref, (list, tuple)):
                    try:
                        idx = ref.index(obj)
                        print(f"{' ' * (depth+2)}(via {type(ref).__name__} index: {idx})")
                    except (ValueError, TypeError):
                        print(f"{' ' * (depth+2)}(as part of a larger structure)")
                
                # Look for owners of this container
                if isinstance(ref, dict):
                    owners = [o for o in gc.get_referrers(ref) 
                             if hasattr(o, '__dict__') and o.__dict__ is ref]
                    if owners:
                        owner = owners[0]
                        print(f"{' ' * (depth+2)}(dict belongs to: {type(owner).__name__} id:{id(owner)})")
                
                # Continue recursion with increased depth
                find_path_to_referrers(ref, current_path, depth + 1, visited)
                
        # Start the recursive search
        find_path_to_referrers(self.target_obj)


from typing import Type, List
from .data_structures import BaseModel
from .response_validator_abc import ResponseValidatorABC


class ResponseValidatorFactory:
    """Factory class to create a response validator for a question."""

    def __init__(self, question):
        self.question = question

    @property
    def response_model(self) -> Type["BaseModel"]:
        if self.question._response_model is not None:
            return self.question._response_model
        else:
            return self.question.create_response_model()

    @property
    def response_validator(self) -> "ResponseValidatorABC":
        """Return the response validator."""
        params = {}
        params.update({"response_model": self.question.response_model})
        params.update({k: getattr(self.question, k) for k in self.validator_parameters})
        params.update({"exception_to_throw": getattr(self.question, "exception_to_throw", None)})
        params.update({"override_answer": getattr(self.question, "override_answer", None)})
        return self.question.response_validator_class(**params)

    @property
    def validator_parameters(self) -> List[str]:
        """Return the parameters required for the response validator."""
        return self.question.response_validator_class.required_params



"""
The tokens module provides functionality for tracking and analyzing token usage in EDSL.

This module implements classes for tracking and reporting token usage across various
components of EDSL, particularly for language model calls. It supports aggregation,
cost calculation, and reporting of token usage metrics.

Key components:
1. TokenUsage - Tracks prompt and completion tokens for a single operation
2. InterviewTokenUsage - Aggregates token usage across an entire interview
3. Exception classes for handling token-related errors

The token tracking system helps with:
- Cost estimation and billing
- Resource utilization analysis
- Cache effectiveness measurement
- API quota management
"""

from .token_usage import TokenUsage
from .interview_token_usage import InterviewTokenUsage
from .exceptions import TokenError, TokenUsageError, TokenCostError

__all__ = [
    "TokenUsage", 
    "InterviewTokenUsage",
    "TokenError",
    "TokenUsageError", 
    "TokenCostError"
]


"""
Progress bar management for asynchronous job execution.

This module provides a context manager for handling progress bar setup and thread
management during job execution. It coordinates the display and updating of progress
bars, particularly for remote tracking via the Expected Parrot API.
"""

import threading
import warnings

from ..coop import Coop
from .jobs_runner_status import JobsRunnerStatus


class ProgressBarManager:
    """Context manager for handling progress bar setup and thread management.
    
    This class manages the progress bar display and updating during job execution,
    particularly for remote tracking via the Expected Parrot API.
    
    It handles:
    1. Setting up a status tracking object
    2. Creating and managing a background thread for progress updates
    3. Properly cleaning up resources when execution completes
    """
    
    def __init__(self, jobs, run_config, parameters):
        self.parameters = parameters
        self.jobs = jobs
        
        # Set up progress tracking
        coop = Coop()
        endpoint_url = coop.get_progress_bar_url()
        
        # Set up jobs status object
        params = {
            "jobs": jobs,
            "n": parameters.n,
            "endpoint_url": endpoint_url,
            "job_uuid": parameters.job_uuid,
        }
        
        # If the jobs_runner_status is already set, use it directly
        if run_config.environment.jobs_runner_status is not None:
            self.jobs_runner_status = run_config.environment.jobs_runner_status
        else:
            # Otherwise create a new one
            self.jobs_runner_status = JobsRunnerStatus(**params)
        
        # Store on run_config for use by other components
        run_config.environment.jobs_runner_status = self.jobs_runner_status
        
        self.progress_thread = None
        self.stop_event = threading.Event()

    def __enter__(self):
        if self.parameters.progress_bar and self.jobs_runner_status.has_ep_api_key():
            self.jobs_runner_status.setup()
            self.progress_thread = threading.Thread(
                target=self._run_progress_bar,
                args=(self.stop_event, self.jobs_runner_status)
            )
            self.progress_thread.start()
        elif self.parameters.progress_bar:
            warnings.warn(
                "You need an Expected Parrot API key to view job progress bars."
            )
        return self.stop_event

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_event.set()
        if self.progress_thread is not None:
            self.progress_thread.join()

    @staticmethod
    def _run_progress_bar(stop_event, jobs_runner_status):
        """Runs the progress bar in a separate thread."""
        jobs_runner_status.update_progress(stop_event)


from typing import (
    Protocol,
    Optional,
    TYPE_CHECKING,
    Sequence,
    Union,
    Literal,
)

if TYPE_CHECKING:
    from ..dataset import Dataset

from .table_data_class import TableData

from .table_renderers import DataTablesRenderer, PandasStyleRenderer, RichRenderer

Row = Sequence[Union[str, int, float, bool, None]]
TableFormat = Literal[
    "grid", "simple", "pipe", "orgtbl", "rst", "mediawiki", "html", "latex", "rich"
]


class TableRenderer(Protocol):
    """Table renderer protocol"""

    def render_html(self, table_data: TableData) -> str:
        pass


# Modified TableDisplay class
class TableDisplay:
    def __init__(
        self,
        headers: Sequence[str],
        data: Sequence[Row],
        tablefmt: Optional[TableFormat] = None,
        raw_data_set: "Dataset" = None,
        renderer_class: Optional[TableRenderer] = None,
    ):
        assert len(headers) == len(data[0])  # Check if headers and data are consistent

        self.headers = headers
        self.data = data
        self.tablefmt = tablefmt
        self.raw_data_set = raw_data_set

        self.renderer_class = renderer_class or PandasStyleRenderer

        # Handle printing parameters from raw_data_set
        if hasattr(raw_data_set, "print_parameters"):
            self.printing_parameters = (
                raw_data_set.print_parameters if raw_data_set.print_parameters else {}
            )
        else:
            self.printing_parameters = {}

    def _repr_html_(self) -> str:
        """
        HTML representation for Jupyter/Colab notebooks.

        The primary path uses the configured `renderer_class` to build an HTML
        string.  Unfortunately, in shared or long-running notebook runtimes it
        is not uncommon for binary dependencies (NumPy, Pandas, etc.) to get
        into an incompatible state, raising import-time errors that would
        otherwise bubble up to the notebook and obscure the actual table
        output.  To make the developer experience smoother we catch *any*
        exception, log/annotate it, and fall back to a plain-text rendering via
        `tabulate`, wrapped in a <pre> block so at least a readable table is
        shown.
        """
        try:
            table_data = TableData(
                headers=self.headers,
                data=self.data,
                parameters=self.printing_parameters,
                raw_data_set=self.raw_data_set,
            )
            return self.renderer_class(table_data).render_html()
        except Exception as exc:  # pragma: no cover
            # --- graceful degradation -------------------------------------------------
            try:
                from tabulate import tabulate

                plain = tabulate(
                    self.data,
                    headers=self.headers,
                    tablefmt=self.tablefmt or "simple",
                )
            except Exception:
                # Even `tabulate` failed – resort to the default __repr__.
                plain = super().__repr__() if hasattr(super(), "__repr__") else str(self.data)

            # Escape HTML-sensitive chars so the browser renders plain text.
            import html

            safe_plain = html.escape(plain)
            return f"<pre>{safe_plain}\n\n[TableDisplay fallback – original error: {exc}]</pre>"

    def __repr__(self):
        # If rich format is requested, use RichRenderer
        if self.tablefmt == "rich":
            table_data = TableData(
                headers=self.headers,
                data=self.data,
                parameters=self.printing_parameters,
                raw_data_set=self.raw_data_set,
            )
            RichRenderer(table_data).render_terminal()
            return ""  # Return empty string since the table is already printed
        else:
            # Fall back to tabulate for other formats
            from tabulate import tabulate
            return tabulate(self.data, headers=self.headers, tablefmt=self.tablefmt)

    @classmethod
    def from_dictionary(
        cls,
        dictionary: dict,
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers = list(dictionary.keys())
        data = [list(dictionary.values())]
        return cls(headers, data, tablefmt, renderer_class=renderer)

    @classmethod
    def from_dictionary_wide(
        cls,
        dictionary: dict,
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers = ["key", "value"]
        data = [[k, v] for k, v in dictionary.items()]
        return cls(headers, data, tablefmt, renderer_class=renderer)

    @classmethod
    def from_dataset(
        cls,
        dataset: "Dataset",
        tablefmt: Optional[TableFormat] = None,
        renderer: Optional[TableRenderer] = None,
    ) -> "TableDisplay":
        headers, data = dataset._tabular()
        return cls(headers, data, tablefmt, dataset, renderer_class=renderer)

    def long(self) -> "TableDisplay":
        """Convert to long format"""
        new_header = ["row", "key", "value"]
        new_data = []
        for index, row in enumerate(self.data):
            new_data.extend([[index, k, v] for k, v in zip(self.headers, row)])
        return TableDisplay(
            new_header, new_data, self.tablefmt, renderer_class=self.renderer_class
        )


# Example usage:
if __name__ == "__main__":
    headers = ["Name", "Age", "City"]
    data = [["John", 30, "New York"], ["Jane", 25, "London"]]

    # Using default (Pandas) renderer
    table1 = TableDisplay(headers, data)

    # Using DataTables renderer
    table2 = TableDisplay(headers, data, renderer=DataTablesRenderer())



"""
Custom exceptions for the notebooks module.
"""

from ..base import BaseException


class NotebookError(BaseException):
    """
    Base exception class for all notebook-related errors.
    
    This is the parent class for all exceptions related to notebook
    operations, including creation, validation, and conversion.
    """
    relevant_doc = "https://docs.expectedparrot.com/"


class NotebookValueError(NotebookError):
    """
    Exception raised when an invalid value is provided to a notebook method.
    
    This exception occurs when attempting to create or modify a notebook
    with invalid values, such as:
    - Invalid data format
    - Incompatible notebook contents
    
    Examples:
        ```python
        # Attempting to create a notebook with invalid data
        notebook = Notebook(data=invalid_data)  # Raises NotebookValueError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/"
    

class NotebookFormatError(NotebookError):
    """
    Exception raised when a notebook's format is invalid.
    
    This exception occurs when the notebook structure does not conform
    to the expected Jupyter Notebook format.
    
    Examples:
        ```python
        # Attempting to load a notebook with invalid format
        notebook = Notebook.from_dict(invalid_dict)  # Raises NotebookFormatError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/"


class NotebookConversionError(NotebookError):
    """
    Exception raised when a notebook conversion fails.
    
    This exception occurs when attempting to convert a notebook to another
    format (like LaTeX) and the conversion process fails.
    
    Examples:
        ```python
        # Attempting to convert a notebook to LaTeX with invalid options
        notebook.to_latex(invalid_options)  # Raises NotebookConversionError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/"


class NotebookEnvironmentError(NotebookError):
    """
    Exception raised when the notebook environment is not supportable.
    
    This exception occurs when attempting to create a notebook in an 
    environment that doesn't provide required context, such as when
    trying to create a notebook from within itself outside of VS Code.
    
    Examples:
        ```python
        # Attempting to create a notebook from within itself in an unsupported IDE
        notebook = Notebook()  # Raises NotebookEnvironmentError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/"


import tempfile
from ..file_methods import FileMethods


class Mp4Methods(FileMethods):
    """
    Handler for MP4 video files.
    
    This class provides methods to handle MP4 video files in both notebook
    and system environments, including viewing and creating example videos.
    """
    suffix = "mp4"

    def view_system(self):
        """
        Open the MP4 file with the system's default video player.
        """
        import os
        import subprocess

        if os.path.exists(self.path):
            try:
                if (os_name := os.name) == "posix":
                    subprocess.run(["open", self.path], check=True)  # macOS
                elif os_name == "nt":
                    os.startfile(self.path)  # Windows
                else:
                    subprocess.run(["xdg-open", self.path], check=True)  # Linux
            except Exception as e:
                print(f"Error opening MP4: {e}")
        else:
            print("MP4 file was not found.")

    def view_notebook(self):
        """
        Display the MP4 video in a Jupyter notebook using IPython's HTML display.
        """
        from IPython.display import HTML, display
        import base64
        
        # Read the video file and encode it as base64
        with open(self.path, 'rb') as f:
            video_data = f.read()
        
        video_base64 = base64.b64encode(video_data).decode('utf-8')
        
        # Create an HTML5 video element with the base64-encoded video
        video_html = f"""
        <video width="640" height="360" controls>
            <source src="data:video/mp4;base64,{video_base64}" type="video/mp4">
            Your browser does not support the video tag.
        </video>
        """
        
        display(HTML(video_html))

    def extract_text(self):
        """
        Extract text from the video using subtitle extraction (if available).
        Currently returns a message that text extraction is not supported for videos.
        
        Returns:
            str: Message indicating text extraction is not supported
        """
        return "Text extraction is not supported for video files."

    def example(self):
        """
        Create a simple example MP4 file.
        
        Uses FFmpeg to generate a test video pattern if available, 
        otherwise creates a minimal MP4 header.
        
        Returns:
            str: Path to the created example MP4 file
        """
        import os
        import subprocess
        
        # Create a temporary file for the output
        with tempfile.NamedTemporaryFile(delete=False, suffix=".mp4") as f:
            output_path = f.name
        
        try:
            # Try to use ffmpeg to generate a test pattern video
            subprocess.run(
                [
                    "ffmpeg", "-y", "-f", "lavfi", "-i", "testsrc=duration=5:size=1280x720:rate=30", 
                    "-vcodec", "libx264", "-pix_fmt", "yuv420p", output_path
                ],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            return output_path
        except (subprocess.SubprocessError, FileNotFoundError):
            # If ffmpeg is not available, create a tiny placeholder MP4 file
            # Using a simple empty binary file with the .mp4 extension
            with open(output_path, 'wb') as f:
                # Just write a simple 1KB file with MP4 signature
                f.write(b'\x00\x00\x00\x18\x66\x74\x79\x70\x6D\x70\x34\x32')  # MP4 file signature
                f.write(b'\x00' * 1000)  # Fill with zeros
            
            return output_path


"""
Buckets module for managing rate limits of language model API requests.

This module provides a robust rate-limiting system for language model API calls,
implementing the token bucket algorithm to manage both requests-per-minute and
tokens-per-minute limits. It supports both local (in-process) and remote
(distributed) rate limiting through a client-server architecture.

Key components:
- TokenBucket: Core rate-limiting class implementing the token bucket algorithm
- ModelBuckets: Manages rate limits for a specific language model, containing
  separate buckets for requests and tokens
- BucketCollection: Manages multiple ModelBuckets instances across different
  language model services

The module also includes a FastAPI server implementation (token_bucket_api) and
client (token_bucket_client) for distributed rate limiting scenarios where
multiple processes or machines need to share rate limits.
"""

from .exceptions import (
    BucketError,
    TokenLimitError,
    TokenBucketClientError,
    BucketConfigurationError,
)

from .token_bucket import TokenBucket
from .model_buckets import ModelBuckets
from .token_bucket_client import TokenBucketClient  # Add explicit import for TokenBucketClient

# Import BucketCollection last to avoid circular import issues
from .bucket_collection import BucketCollection

__all__ = [
    "BucketCollection", 
    "ModelBuckets", 
    "TokenBucket",
    "TokenBucketClient",
    "BucketError",
    "TokenLimitError",
    "TokenBucketClientError",
    "BucketConfigurationError",
]



from .instruction import Instruction
from .instruction_collection import InstructionCollection
from .change_instruction import ChangeInstruction
from .instruction_handler import InstructionHandler

__all__ = ["Instruction", "ChangeInstruction", "InstructionCollection", "InstructionHandler"]



"""
Exceptions specific to the interviews module.

This module defines custom exception classes for all interview-related errors
in the EDSL framework, ensuring consistent error handling and user feedback.
"""

from ..base import BaseException


class InterviewError(BaseException):
    """
    Base exception class for all interview-related errors.
    
    This is the parent class for all exceptions related to interview creation,
    execution, and management in the EDSL framework.
    
    Examples:
        ```python
        # Usually not raised directly, but through subclasses
        # For example, when accessing incomplete tasks
        interview.get_completed_task("incomplete_task")  # Would raise InterviewTaskError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/en/latest/interviews.html"


class InterviewTaskError(InterviewError):
    """
    Exception raised when there's an issue with interview tasks.
    
    This exception occurs when:
    - Attempting to access tasks that are not complete
    - Task execution fails due to errors in the model response
    - Task dependencies are not satisfied
    
    Examples:
        ```python
        # Attempting to access an incomplete task
        interview.get_completed_task("incomplete_task")  # Raises InterviewTaskError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/en/latest/interviews.html"


class InterviewStatusError(InterviewError):
    """
    Exception raised when there's an issue with interview status management.
    
    This exception occurs when:
    - Invalid operations are performed on interview status objects
    - Incompatible status objects are combined
    - Status tracking encounters inconsistent states
    
    Examples:
        ```python
        # Attempting to add incompatible status objects
        status_dict + incompatible_object  # Raises InterviewStatusError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/en/latest/interviews.html"


class InterviewTokenError(InterviewError):
    """
    Exception raised when there's an issue with token estimation or usage.
    
    This exception occurs when:
    - Invalid inputs are provided to token estimators
    - Token calculation fails due to unsupported prompt types
    - Token limits are exceeded during an interview
    
    Examples:
        ```python
        # Providing an invalid prompt type to token estimator
        estimator.estimate_tokens(invalid_prompt)  # Raises InterviewTokenError
        ```
    """
    relevant_doc = "https://docs.expectedparrot.com/en/latest/token_usage.html"


import asyncio
import copy
import weakref
from typing import TYPE_CHECKING, Any, Callable, Union

if TYPE_CHECKING:
    from ..invigilators.invigilator_base import InvigilatorBase
    from ..key_management import KeyLookup
    from ..questions import QuestionBase
    from .interview import Interview

from ..data_transfer_models import EDSLResultObjectInput
from ..jobs.fetch_invigilator import FetchInvigilator
from ..language_models.exceptions import LanguageModelNoResponseError
from ..questions.exceptions import QuestionAnswerValidationError
from ..surveys.base import EndOfSurvey
from ..tasks import TaskStatus
from .exception_tracking import InterviewExceptionEntry


class RetryConfig:
    from ..config import CONFIG

    EDSL_BACKOFF_START_SEC = float(CONFIG.get("EDSL_BACKOFF_START_SEC"))
    EDSL_BACKOFF_MAX_SEC = float(CONFIG.get("EDSL_BACKOFF_MAX_SEC"))
    EDSL_MAX_ATTEMPTS = int(CONFIG.get("EDSL_MAX_ATTEMPTS"))


class SkipHandler:
    def __init__(self, interview: "Interview"):
        # Store a weak reference to the interview
        self._interview_ref = weakref.ref(interview)

        # Cache only the skip function which doesn't maintain a reference to the interview
        try:
            self.skip_function: Callable = (
                interview.survey.rule_collection.skip_question_before_running
            )
        except (AttributeError, KeyError):
            # Fallback for test environments
            self.skip_function = lambda *args: False

    @property
    def interview(self):
        """Access the interview via weak reference if it still exists."""
        interview = self._interview_ref()
        if interview is None:
            raise RuntimeError("Interview has been garbage collected")
        return interview

    @property
    def _to_index(self):
        return self.interview.to_index

    @property
    def _survey(self):
        return self.interview.survey

    @property
    def _answers(self):
        return self.interview.answers

    @property
    def _scenario(self):
        return self.interview.scenario

    @property
    def _agent_traits(self):
        try:
            return self.interview.agent["traits"]
        except (AttributeError, KeyError):
            return {}

    @property
    def _skip_flags(self):
        return self.interview.skip_flags

    def should_skip(self, current_question: "QuestionBase") -> bool:
        """Determine if the current question should be skipped."""
        current_question_index = self._to_index[current_question.question_name]

        # Handle ScenarioList case - convert to dict first
        scenario_dict = {}
        if hasattr(self._scenario, "items"):
            # Handle standard dict scenario
            scenario_dict = self._scenario
        else:
            # Handle ScenarioList or other scenario object
            # Access as a dict if possible, otherwise try to convert
            scenario_dict = (
                dict(self._scenario) if hasattr(self._scenario, "__iter__") else {}
            )

        combined_answers = dict(self._answers)
        combined_answers.update(scenario_dict)
        combined_answers.update(self._agent_traits)

        return self.skip_function(current_question_index, combined_answers)

    def _current_info_env(self) -> dict[str, Any]:
        """
        - The current answers are "generated_tokens" and "comment"
        - The scenario should have "scenario." added to the keys
        - The agent traits should have "agent." added to the keys
        """
        # Process answers dictionary
        processed_answers = {}
        for key, value in self._answers.items():
            if key.endswith("_generated_tokens"):
                base_name = key.replace("_generated_tokens", "")
                processed_answers[f"{base_name}.generated_tokens"] = value
            elif key.endswith("_comment"):
                base_name = key.replace("_comment", "")
                processed_answers[f"{base_name}.comment"] = value
            else:
                # Regular answer
                processed_answers[f"{key}.answer"] = value

        # Process scenario dictionary
        processed_scenario = {f"scenario.{k}": v for k, v in self._scenario.items()}

        # Process agent traits
        processed_agent = {f"agent.{k}": v for k, v in self._agent_traits.items()}

        return processed_answers | processed_scenario | processed_agent

    def cancel_skipped_questions(self, current_question: "QuestionBase") -> None:
        """Cancel the tasks for questions that should be skipped."""
        current_question_index: int = self._to_index[current_question.question_name]
        answers = self._current_info_env()

        # Get the index of the next question, which could also be the end of the survey
        next_question: Union[
            int, EndOfSurvey
        ] = self._survey.rule_collection.next_question(
            q_now=current_question_index,
            answers=answers,
        )

        def cancel_between(start, end):
            """Cancel the tasks for questions between the start and end indices."""
            for i in range(start, end):
                # print(f"Cancelling task {i}")
                # self.interview.tasks[i].cancel()
                # self.interview.tasks[i].set_result("skipped")
                interview = self._interview_ref()
                if interview is not None:
                    interview.skip_flags[self._survey.questions[i].question_name] = True
                else:
                    # If interview is gone, there's nothing to skip anymore
                    return

        if (next_question_index := next_question.next_q) == EndOfSurvey:
            cancel_between(current_question_index + 1, len(self._survey.questions))
            return

        if next_question_index > (current_question_index + 1):
            cancel_between(current_question_index + 1, next_question_index)


class AnswerQuestionFunctionConstructor:
    """Constructs a function that answers a question and records the answer."""

    def __init__(self, interview: "Interview", key_lookup: "KeyLookup"):
        # Store a weak reference to the interview
        self._interview_ref = weakref.ref(interview)
        self.key_lookup = key_lookup

        # Store configuration settings that won't change during lifecycle
        self._raise_validation_errors = getattr(
            interview, "raise_validation_errors", False
        )
        self._stop_on_exception = getattr(interview, "stop_on_exception", False)

        self.had_language_model_no_response_error: bool = False

        # Initialize fetch invigilator with the interview - this should use weakref internally
        self.invigilator_fetcher = FetchInvigilator(
            interview, key_lookup=self.key_lookup
        )

        # In our test environment, we might not be able to create the SkipHandler
        # because example Interview might not have all required attributes
        # So we'll initialize it conditionally
        if hasattr(interview, "skip_flags"):
            self.skip_handler = SkipHandler(interview)
        else:
            self.skip_handler = None

    @property
    def interview(self):
        """Access the interview via weak reference if it still exists."""
        interview = self._interview_ref()
        if interview is None:
            raise RuntimeError("Interview has been garbage collected")
        return interview

    @property
    def _answers(self):
        return self.interview.answers

    @property
    def _exceptions(self):
        return self.interview.exceptions

    @property
    def _to_index(self):
        return self.interview.to_index

    @property
    def _skip_flags(self):
        if hasattr(self.interview, "skip_flags"):
            return self.interview.skip_flags
        return {}

    def _handle_exception(
        self, e: Exception, invigilator: "InvigilatorBase", task=None
    ):
        """Handle an exception that occurred while answering a question."""
        interview = self._interview_ref()
        if interview is None:
            # If interview is gone, we can't really handle the exception properly
            # Just raise it to the caller
            raise e

        # Copy to freeze the answers here for logging
        answers = copy.copy(self._answers)

        exception_entry = InterviewExceptionEntry(
            exception=e,
            invigilator=invigilator,
            answers=answers,
        )

        if task:
            task.task_status = TaskStatus.FAILED

        # Add to exceptions - need to use the interview reference here
        interview.exceptions.add(invigilator.question.question_name, exception_entry)

        # Check if we should raise validation errors
        if self._raise_validation_errors and isinstance(
            e, QuestionAnswerValidationError
        ):
            raise e

        # Check if we should stop on exception
        if self._stop_on_exception:
            raise e

    def __call__(self):
        return self.answer_question_and_record_task

    async def answer_question_and_record_task(
        self,
        *,
        question: "QuestionBase",
        task=None,
    ) -> "EDSLResultObjectInput":
        from tenacity import (
            RetryError,
            retry,
            retry_if_exception_type,
            stop_after_attempt,
            wait_exponential,
        )

        @retry(
            stop=stop_after_attempt(RetryConfig.EDSL_MAX_ATTEMPTS),
            wait=wait_exponential(
                multiplier=RetryConfig.EDSL_BACKOFF_START_SEC,
                max=RetryConfig.EDSL_BACKOFF_MAX_SEC,
            ),
            retry=retry_if_exception_type(LanguageModelNoResponseError),
            reraise=True,
        )
        async def attempt_answer():
            # Get a reference to the interview (may be None if it's been garbage collected)
            interview = self._interview_ref()

            # Get the invigilator for this question
            invigilator = self.invigilator_fetcher(question)

            # Check if interview still exists
            if interview is None:
                # If interview is gone, we can't really process this question
                # Return a failure result
                return invigilator.get_failed_task_result(
                    failure_reason="Interview has been garbage collected."
                )

            # Check if question should be skipped - use cached skip_flags if available
            skip_flags = getattr(self, "_skip_flags", None) or interview.skip_flags
            if skip_flags.get(question.question_name, False):
                return invigilator.get_failed_task_result(
                    failure_reason="Question skipped."
                )

            if self.skip_handler and self.skip_handler.should_skip(question):
                return invigilator.get_failed_task_result(
                    failure_reason="Question skipped."
                )

            had_language_model_no_response_error = False
            try:
                response: EDSLResultObjectInput = (
                    await invigilator.async_answer_question()
                )
                if response.validated:
                    # Re-check if interview exists before updating it
                    interview = self._interview_ref()
                    if interview is not None:
                        interview.answers.add_answer(
                            response=response, question=question
                        )
                        if self.skip_handler:
                            self.skip_handler.cancel_skipped_questions(question)
                else:
                    if (
                        hasattr(response, "exception_occurred")
                        and response.exception_occurred
                    ):
                        raise response.exception_occurred

            except QuestionAnswerValidationError as e:
                self._handle_exception(e, invigilator, task)
                return invigilator.get_failed_task_result(
                    failure_reason="Question answer validation failed."
                )

            except asyncio.TimeoutError as e:
                self._handle_exception(e, invigilator, task)
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model timed out for question '{question.question_name}.'"
                )

            except Exception as e:
                self._handle_exception(e, invigilator, task)

            if "response" not in locals():
                had_language_model_no_response_error = True
                raise LanguageModelNoResponseError(
                    f"Language model did not return a response for question '{question.question_name}.'"
                )

            # Re-check if interview exists before accessing exceptions
            interview = self._interview_ref()
            if (
                interview is not None
                and question.question_name in interview.exceptions
                and had_language_model_no_response_error
            ):
                interview.exceptions.record_fixed_question(question.question_name)

            return response

        try:
            out = await attempt_answer()
            return out
        except RetryError as retry_error:
            original_error = retry_error.last_attempt.exception()
            self._handle_exception(
                original_error, self.invigilator_fetcher(question), task
            )
            raise original_error



from typing import TYPE_CHECKING
from typing import Callable
from edsl.agents import Agent
#from edsl.surveys import Survey

if TYPE_CHECKING:
    from edsl.surveys import Survey
    from edsl.results import Results
    from edsl.questions import QuestionBase

class Simulator:
    def __init__(self, survey: "Survey"):
        self.survey = survey

    @classmethod
    def random_survey(cls):
        """Create a random survey."""
        from edsl.questions import QuestionMultipleChoice, QuestionFreeText
        from random import choice
        from edsl.surveys import Survey

        num_questions = 10
        questions = []
        for i in range(num_questions):
            if choice([True, False]):
                q = QuestionMultipleChoice(
                    question_text="nothing",
                    question_name="q_" + str(i),
                    question_options=list(range(3)),
                )
                questions.append(q)
            else:
                questions.append(
                    QuestionFreeText(
                        question_text="nothing", question_name="q_" + str(i)
                    )
                )
        s = Survey(questions)
        start_index = choice(range(num_questions - 1))
        end_index = choice(range(start_index + 1, 10))
        s = s.add_rule(f"q_{start_index}", "True", f"q_{end_index}")
        question_to_delete = choice(range(num_questions))
        s.delete_question(f"q_{question_to_delete}")
        return s

    def simulate(self) -> dict:
        """Simulate the survey and return the answers."""
        i = self.survey.gen_path_through_survey()
        q = next(i)
        num_passes = 0
        while True:
            num_passes += 1
            try:
                answer = q._simulate_answer()
                q = i.send({q.question_name: answer["answer"]})
            except StopIteration:
                break

            if num_passes > 100:
                print("Too many passes.")
                from .exceptions import SurveyError
                raise SurveyError("Too many passes.")
        return self.survey.answers

    def create_agent(self) -> "Agent":
        """Create an agent from the simulated answers."""
        answers_dict = self.survey.simulate()
        from edsl.agents import Agent

        def construct_answer_dict_function(traits: dict) -> Callable:
            def func(self, question: "QuestionBase", scenario=None):
                return traits.get(question.question_name, None)

            return func

        return Agent(traits=answers_dict).add_direct_question_answering_method(
            construct_answer_dict_function(answers_dict)
        )

    def simulate_results(self) -> "Results":
        """Simulate the survey and return the results."""
        a = self.create_agent()
        return self.survey.by([a]).run()



"""
Core display functionality that abstracts IPython.display.

This module provides wrapper classes and functions that abstract the IPython.display
functionality to enable potential future replacement with alternative implementations.
"""

try:
    from IPython.display import display as ipython_display
    from IPython.display import HTML as IPythonHTML
    from IPython.display import FileLink as IPythonFileLink
    from IPython.display import IFrame as IPythonIFrame
    _IPYTHON_AVAILABLE = True
except ImportError:
    _IPYTHON_AVAILABLE = False


def is_notebook_environment():
    """
    Check if code is running in a Jupyter notebook or similar interactive environment.
    
    Returns:
        bool: True if running in a notebook environment, False otherwise
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "Shell":  # Google Colab's shell class
            import sys
            if "google.colab" in sys.modules:
                return True  # Running in Google Colab
            return False
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type
    except (NameError, ImportError):
        return False  # Probably standard Python interpreter


class HTML:
    """
    Wrapper for IPython's HTML display class.
    
    This class provides the same functionality as IPython.display.HTML but can be
    extended or replaced with alternative implementations.
    """
    def __init__(self, data=None, metadata=None, **kwargs):
        self.data = data
        self.metadata = metadata
        self.kwargs = kwargs
        
        if _IPYTHON_AVAILABLE and is_notebook_environment():
            self._ipython_html = IPythonHTML(data, metadata, **kwargs)
        else:
            self._ipython_html = None

    def __repr__(self):
        """Return a string representation of the HTML object."""
        return f"{self.data}"

    def _repr_html_(self):
        """Return HTML representation of the object."""
        if self._ipython_html:
            return self._ipython_html._repr_html_()
        return self.data


class FileLink:
    """
    Wrapper for IPython's FileLink display class.
    
    This class provides the same functionality as IPython.display.FileLink but can be
    extended or replaced with alternative implementations.
    """
    def __init__(self, path, url_prefix='', result_html_prefix='', result_html_suffix='', **kwargs):
        self.path = path
        self.url_prefix = url_prefix
        self.result_html_prefix = result_html_prefix 
        self.result_html_suffix = result_html_suffix
        self.kwargs = kwargs
        
        if _IPYTHON_AVAILABLE and is_notebook_environment():
            self._ipython_filelink = IPythonFileLink(
                path, 
                url_prefix, 
                result_html_prefix, 
                result_html_suffix,
                **kwargs
            )
        else:
            self._ipython_filelink = None

    def _repr_html_(self):
        """Return HTML representation of the file link."""
        if self._ipython_filelink:
            return self._ipython_filelink._repr_html_()
        return f'<a href="{self.url_prefix}{self.path}" target="_blank">{self.path}</a>'


class IFrame:
    """
    Wrapper for IPython's IFrame display class.
    
    This class provides the same functionality as IPython.display.IFrame but can be
    extended or replaced with alternative implementations.
    """
    def __init__(self, src, width, height, **kwargs):
        self.src = src
        self.width = width
        self.height = height
        self.kwargs = kwargs
        
        if _IPYTHON_AVAILABLE and is_notebook_environment():
            self._ipython_iframe = IPythonIFrame(src, width, height, **kwargs)
        else:
            self._ipython_iframe = None

    def _repr_html_(self):
        """Return HTML representation of the iframe."""
        if self._ipython_iframe:
            return self._ipython_iframe._repr_html_()
        return f'<iframe src="{self.src}" width="{self.width}" height="{self.height}"></iframe>'


def display(obj, *args, **kwargs):
    """
    Display an object in the frontend.
    
    Wrapper around IPython.display.display that can be extended or replaced with
    alternative implementations.
    
    Args:
        obj: The object to display
        *args: Additional objects to display
        **kwargs: Additional keyword arguments for display
    """
    if _IPYTHON_AVAILABLE and is_notebook_environment():
        ipython_display(obj, *args, **kwargs)
    else:
        # Fallback behavior when not in notebook environment
        if hasattr(obj, '_repr_html_'):
            print("HTML representation available, but not in notebook environment.")
        print(repr(obj))
        for arg in args:
            print(repr(arg))